# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import datetime

import attr
from attrs_strict import AttributeTypeError
from hypothesis import given
from hypothesis.strategies import binary
import pytest

from swh.model.model import (
    Content, SkippedContent, Directory, Revision, Release, Snapshot,
    Origin, Timestamp, TimestampWithTimezone,
    MissingData, Person
)
from swh.model.hashutil import hash_to_bytes, MultiHash
import swh.model.hypothesis_strategies as strategies
from swh.model.identifiers import (
    directory_identifier, revision_identifier, release_identifier,
    snapshot_identifier
)
from swh.model.tests.test_identifiers import (
    directory_example, revision_example, release_example, snapshot_example
)


@given(strategies.objects())
def test_todict_inverse_fromdict(objtype_and_obj):
    (obj_type, obj) = objtype_and_obj

    if obj_type in ('origin', 'origin_visit'):
        return

    obj_as_dict = obj.to_dict()
    obj_as_dict_copy = copy.deepcopy(obj_as_dict)

    # Check the composition of to_dict and from_dict is the identity
    assert obj == type(obj).from_dict(obj_as_dict)

    # Check from_dict() does not change the input dict
    assert obj_as_dict == obj_as_dict_copy

    # Check the composition of from_dict and to_dict is the identity
    assert obj_as_dict == type(obj).from_dict(obj_as_dict).to_dict()


@given(strategies.origins())
def test_todict_origins(origin):
    obj = origin.to_dict()

    assert 'type' not in obj
    assert type(origin)(url=origin.url) == type(origin).from_dict(obj)


@given(strategies.origin_visits())
def test_todict_origin_visits(origin_visit):
    obj = origin_visit.to_dict()

    assert origin_visit == type(origin_visit).from_dict(obj)


@given(strategies.origin_visit_updates())
def test_todict_origin_visit_updates(origin_visit_update):
    obj = origin_visit_update.to_dict()

    assert origin_visit_update == type(origin_visit_update).from_dict(obj)


# Timestamp

@given(strategies.timestamps())
def test_timestamps_strategy(timestamp):
    attr.validate(timestamp)


def test_timestamp_seconds():
    attr.validate(Timestamp(seconds=0, microseconds=0))
    with pytest.raises(AttributeTypeError):
        Timestamp(seconds='0', microseconds=0)

    attr.validate(Timestamp(seconds=2**63-1, microseconds=0))
    with pytest.raises(ValueError):
        Timestamp(seconds=2**63, microseconds=0)

    attr.validate(Timestamp(seconds=-2**63, microseconds=0))
    with pytest.raises(ValueError):
        Timestamp(seconds=-2**63-1, microseconds=0)


def test_timestamp_microseconds():
    attr.validate(Timestamp(seconds=0, microseconds=0))
    with pytest.raises(AttributeTypeError):
        Timestamp(seconds=0, microseconds='0')

    attr.validate(Timestamp(seconds=0, microseconds=10**6-1))
    with pytest.raises(ValueError):
        Timestamp(seconds=0, microseconds=10**6)

    with pytest.raises(ValueError):
        Timestamp(seconds=0, microseconds=-1)


def test_timestamp_from_dict():
    assert Timestamp.from_dict({'seconds': 10, 'microseconds': 5})

    with pytest.raises(AttributeTypeError):
        Timestamp.from_dict({'seconds': '10', 'microseconds': 5})

    with pytest.raises(AttributeTypeError):
        Timestamp.from_dict({'seconds': 10, 'microseconds': '5'})
    with pytest.raises(ValueError):
        Timestamp.from_dict({'seconds': 0, 'microseconds': -1})

    Timestamp.from_dict({'seconds': 0, 'microseconds': 10**6 - 1})
    with pytest.raises(ValueError):
        Timestamp.from_dict({'seconds': 0, 'microseconds': 10**6})


# TimestampWithTimezone

def test_timestampwithtimezone():
    ts = Timestamp(seconds=0, microseconds=0)
    tstz = TimestampWithTimezone(
        timestamp=ts,
        offset=0,
        negative_utc=False)
    attr.validate(tstz)
    assert tstz.negative_utc is False

    attr.validate(TimestampWithTimezone(
        timestamp=ts,
        offset=10,
        negative_utc=False))

    attr.validate(TimestampWithTimezone(
        timestamp=ts,
        offset=-10,
        negative_utc=False))

    tstz = TimestampWithTimezone(
        timestamp=ts,
        offset=0,
        negative_utc=True)
    attr.validate(tstz)
    assert tstz.negative_utc is True

    with pytest.raises(AttributeTypeError):
        TimestampWithTimezone(
            timestamp=datetime.datetime.now(),
            offset=0,
            negative_utc=False)

    with pytest.raises(AttributeTypeError):
        TimestampWithTimezone(
            timestamp=ts,
            offset='0',
            negative_utc=False)

    with pytest.raises(AttributeTypeError):
        TimestampWithTimezone(
            timestamp=ts,
            offset=1.0,
            negative_utc=False)

    with pytest.raises(AttributeTypeError):
        TimestampWithTimezone(
            timestamp=ts,
            offset=1,
            negative_utc=0)

    with pytest.raises(ValueError):
        TimestampWithTimezone(
            timestamp=ts,
            offset=1,
            negative_utc=True)

    with pytest.raises(ValueError):
        TimestampWithTimezone(
            timestamp=ts,
            offset=-1,
            negative_utc=True)


def test_timestampwithtimezone_from_datetime():
    tz = datetime.timezone(datetime.timedelta(minutes=+60))
    date = datetime.datetime(
        2020, 2, 27, 14, 39, 19, tzinfo=tz)

    tstz = TimestampWithTimezone.from_datetime(date)

    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=1582810759,
            microseconds=0,
        ),
        offset=60,
        negative_utc=False,
    )


def test_timestampwithtimezone_from_iso8601():
    date = '2020-02-27 14:39:19.123456+0100'

    tstz = TimestampWithTimezone.from_iso8601(date)

    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=1582810759,
            microseconds=123456,
        ),
        offset=60,
        negative_utc=False,
    )


def test_timestampwithtimezone_from_iso8601_negative_utc():
    date = '2020-02-27 13:39:19-0000'

    tstz = TimestampWithTimezone.from_iso8601(date)

    assert tstz == TimestampWithTimezone(
        timestamp=Timestamp(
            seconds=1582810759,
            microseconds=0,
        ),
        offset=0,
        negative_utc=True,
    )


def test_person_from_fullname():
    """The author should have name, email and fullname filled.

    """
    actual_person = Person.from_fullname(b'tony <ynot@dagobah>')
    assert actual_person == Person(
        fullname=b'tony <ynot@dagobah>',
        name=b'tony',
        email=b'ynot@dagobah',
    )


def test_person_from_fullname_no_email():
    """The author and fullname should be the same as the input (author).

    """
    actual_person = Person.from_fullname(b'tony')
    assert actual_person == Person(
        fullname=b'tony',
        name=b'tony',
        email=None,
    )


def test_person_from_fullname_empty_person():
    """Empty person has only its fullname filled with the empty
    byte-string.

    """
    actual_person = Person.from_fullname(b'')
    assert actual_person == Person(
        fullname=b'',
        name=None,
        email=None,
    )


def test_git_author_line_to_author():
    # edge case out of the way
    with pytest.raises(TypeError):
        Person.from_fullname(None)

    tests = {
        b'a <b@c.com>': Person(
            name=b'a',
            email=b'b@c.com',
            fullname=b'a <b@c.com>',
        ),
        b'<foo@bar.com>': Person(
            name=None,
            email=b'foo@bar.com',
            fullname=b'<foo@bar.com>',
        ),
        b'malformed <email': Person(
            name=b'malformed',
            email=b'email',
            fullname=b'malformed <email'
        ),
        b'malformed <"<br"@ckets>': Person(
            name=b'malformed',
            email=b'"<br"@ckets',
            fullname=b'malformed <"<br"@ckets>',
        ),
        b'trailing <sp@c.e> ': Person(
            name=b'trailing',
            email=b'sp@c.e',
            fullname=b'trailing <sp@c.e> ',
        ),
        b'no<sp@c.e>': Person(
            name=b'no',
            email=b'sp@c.e',
            fullname=b'no<sp@c.e>',
        ),
        b' more   <sp@c.es>': Person(
            name=b'more',
            email=b'sp@c.es',
            fullname=b' more   <sp@c.es>',
        ),
        b' <>': Person(
            name=None,
            email=None,
            fullname=b' <>',
        ),
    }

    for person in sorted(tests):
        expected_person = tests[person]
        assert expected_person == Person.from_fullname(person)


# Content

def test_content_get_hash():
    hashes = dict(
        sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')
    c = Content(length=42, status='visible', **hashes)
    for (hash_name, hash_) in hashes.items():
        assert c.get_hash(hash_name) == hash_


def test_content_hashes():
    hashes = dict(
        sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')
    c = Content(length=42, status='visible', **hashes)
    assert c.hashes() == hashes


def test_content_data():
    c = Content(
        length=42, status='visible', data=b'foo',
        sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')
    assert c.with_data() == c


def test_content_data_missing():
    c = Content(
        length=42, status='visible',
        sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')
    with pytest.raises(MissingData):
        c.with_data()


@given(strategies.present_contents_d())
def test_content_from_dict(content_d):
    c = Content.from_data(**content_d)
    assert c
    assert c.ctime == content_d['ctime']

    content_d2 = c.to_dict()
    c2 = Content.from_dict(content_d2)
    assert c2.ctime == c.ctime


def test_content_from_dict_str_ctime():
    # test with ctime as a string
    n = datetime.datetime(2020, 5, 6, 12, 34)
    content_d = {
        'ctime': n.isoformat(),
        'data': b'',
        'length': 0,
        'sha1': b'\x00',
        'sha256': b'\x00',
        'sha1_git': b'\x00',
        'blake2s256': b'\x00',
        }
    c = Content.from_dict(content_d)
    assert c.ctime == n


@given(binary(max_size=4096))
def test_content_from_data(data):
    c = Content.from_data(data)
    assert c.data == data
    assert c.length == len(data)
    assert c.status == 'visible'
    for key, value in MultiHash.from_data(data).digest().items():
        assert getattr(c, key) == value


@given(binary(max_size=4096))
def test_hidden_content_from_data(data):
    c = Content.from_data(data, status='hidden')
    assert c.data == data
    assert c.length == len(data)
    assert c.status == 'hidden'
    for key, value in MultiHash.from_data(data).digest().items():
        assert getattr(c, key) == value


# SkippedContent

@given(binary(max_size=4096))
def test_skipped_content_from_data(data):
    c = SkippedContent.from_data(data, reason='reason')
    assert c.reason == 'reason'
    assert c.length == len(data)
    assert c.status == 'absent'
    for key, value in MultiHash.from_data(data).digest().items():
        assert getattr(c, key) == value


@given(strategies.skipped_contents_d())
def test_skipped_content_origin_is_str(skipped_content_d):
    assert SkippedContent.from_dict(skipped_content_d)

    skipped_content_d['origin'] = 'http://path/to/origin'
    assert SkippedContent.from_dict(skipped_content_d)

    skipped_content_d['origin'] = Origin(url='http://path/to/origin')
    with pytest.raises(ValueError, match='origin'):
        SkippedContent.from_dict(skipped_content_d)


# ID computation

def test_directory_model_id_computation():
    dir_dict = directory_example.copy()
    del dir_dict['id']

    dir_id = hash_to_bytes(directory_identifier(dir_dict))
    dir_model = Directory.from_dict(dir_dict)
    assert dir_model.id == dir_id


def test_revision_model_id_computation():
    rev_dict = revision_example.copy()
    del rev_dict['id']

    rev_id = hash_to_bytes(revision_identifier(rev_dict))
    rev_model = Revision.from_dict(rev_dict)
    assert rev_model.id == rev_id


def test_revision_model_id_computation_with_no_date():
    """We can have revision with date to None

    """
    rev_dict = revision_example.copy()
    rev_dict['date'] = None
    rev_dict['committer_date'] = None
    del rev_dict['id']

    rev_id = hash_to_bytes(revision_identifier(rev_dict))
    rev_model = Revision.from_dict(rev_dict)
    assert rev_model.date is None
    assert rev_model.committer_date is None
    assert rev_model.id == rev_id


def test_release_model_id_computation():
    rel_dict = release_example.copy()
    del rel_dict['id']

    rel_id = hash_to_bytes(release_identifier(rel_dict))
    rel_model = Release.from_dict(rel_dict)
    assert isinstance(rel_model.date, TimestampWithTimezone)
    assert rel_model.id == hash_to_bytes(rel_id)


def test_snapshot_model_id_computation():
    snp_dict = snapshot_example.copy()
    del snp_dict['id']

    snp_id = hash_to_bytes(snapshot_identifier(snp_dict))
    snp_model = Snapshot.from_dict(snp_dict)
    assert snp_model.id == snp_id



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime
from pytz import all_timezones, timezone
from random import choice, randint, random, shuffle
from typing import List, Dict

from swh.model.hashutil import MultiHash


PROTOCOLS = ['git', 'http', 'https', 'deb', 'svn', 'mock']
DOMAINS = ['example.com', 'some.long.host.name', 'xn--n28h.tld']
PATHS = ['', '/', '/stuff', '/stuff/',
         '/path/to/resource',
         '/path/with/anchor#id=42',
         '/path/with/qargs?q=1&b']
CONTENT_STATUS = ['visible', 'hidden', 'absent']
MAX_DATE = 3e9  # around 2065


def gen_all_origins():
    for protocol in PROTOCOLS:
        for domain in DOMAINS:
            for urlpath in PATHS:
                yield {'url': '%s://%s%s' % (protocol, domain, urlpath)}


ORIGINS = list(gen_all_origins())


def gen_origins(n: int = 100) -> List:
    """Returns a list of n randomly generated origins suitable for using as
    Storage.add_origin() argument.

    """
    origins = ORIGINS[:]
    shuffle(origins)
    return origins[:n]


def gen_content():
    size = randint(1, 10 * 1024)
    data = bytes(randint(0, 255) for i in range(size))
    status = choice(CONTENT_STATUS)
    h = MultiHash.from_data(data)
    ctime = datetime.fromtimestamp(
        random() * MAX_DATE, timezone(choice(all_timezones)))
    content = {'data': data,
               'status': status,
               'length': size,
               'ctime': ctime,
               **h.digest()}
    if status == 'absent':
        content['reason'] = 'why not'
        content['data'] = None
    return content


def gen_contents(n=20) -> List[Dict]:
    """Returns a list of n randomly generated content objects (as dict) suitable
    for using as Storage.content_add() argument.
    """
    return [gen_content() for i in range(n)]



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
#
# The ValidationError code derives from Django, and is available under the
# following license terms:
#
# Copyright (c) Django Software Foundation and individual contributors. All
# rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be
#        used to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

NON_FIELD_ERRORS = '__all__'


class ValidationError(Exception):
    """An error while validating data."""
    def __init__(self, message, code=None, params=None):
        """
        The `message` argument can be a single error, a list of errors, or a
        dictionary that maps field names to lists of errors. What we define as
        an "error" can be either a simple string or an instance of
        ValidationError with its message attribute set, and what we define as
        list or dictionary can be an actual `list` or `dict` or an instance
        of ValidationError with its `error_list` or `error_dict` attribute set.
        """

        super().__init__(message, code, params)

        if isinstance(message, list) and len(message) == 1:
            message = message[0]

        if isinstance(message, ValidationError):
            if hasattr(message, 'error_dict'):
                message = message.error_dict
            # PY2 has a `message` property which is always there so we can't
            # duck-type on it. It was introduced in Python 2.5 and already
            # deprecated in Python 2.6.
            elif not hasattr(message, 'message'):
                message = message.error_list
            else:
                message, code, params = (message.message, message.code,
                                         message.params)

        if isinstance(message, dict):
            self.error_dict = {}
            for field, messages in message.items():
                if not isinstance(messages, ValidationError):
                    messages = ValidationError(messages)
                self.error_dict[field] = messages.error_list

        elif isinstance(message, list):
            self.error_list = []
            for message in message:
                # Normalize plain strings to instances of ValidationError.
                if not isinstance(message, ValidationError):
                    message = ValidationError(message)
                if hasattr(message, 'error_dict'):
                    self.error_list.extend(sum(message.error_dict.values(),
                                               []))
                else:
                    self.error_list.extend(message.error_list)

        else:
            self.message = message
            self.code = code
            self.params = params
            self.error_list = [self]

    @property
    def message_dict(self):
        # Trigger an AttributeError if this ValidationError
        # doesn't have an error_dict.
        getattr(self, 'error_dict')

        return dict(self)

    @property
    def messages(self):
        if hasattr(self, 'error_dict'):
            return sum(dict(self).values(), [])
        return list(self)

    def update_error_dict(self, error_dict):
        if hasattr(self, 'error_dict'):
            for field, error_list in self.error_dict.items():
                error_dict.setdefault(field, []).extend(error_list)
        else:
            error_dict.setdefault(NON_FIELD_ERRORS, []).extend(self.error_list)
        return error_dict

    def __iter__(self):
        if hasattr(self, 'error_dict'):
            for field, errors in self.error_dict.items():
                yield field, list(ValidationError(errors))
        else:
            for error in self.error_list:
                message = error.message
                if error.params:
                    message %= error.params
                yield message

    def __str__(self):
        if hasattr(self, 'error_dict'):
            return repr(dict(self))
        return repr(list(self))

    def __repr__(self):
        return 'ValidationError(%s)' % self



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import attr
from hypothesis import given, settings

from swh.model.hashutil import DEFAULT_ALGORITHMS
from swh.model.hypothesis_strategies import (
    objects, object_dicts, snapshots
)
from swh.model.model import TargetType


target_types = (
    'content', 'directory', 'revision', 'release', 'snapshot', 'alias')


@given(objects())
def test_generation(obj_type_and_obj):
    (obj_type, object_) = obj_type_and_obj
    attr.validate(object_)


def assert_nested_dict(obj):
    """Tests the object is a nested dict and contains no more class
    from swh.model.model."""
    if isinstance(obj, dict):
        for (key, value) in obj.items():
            assert isinstance(key, (str, bytes)), key
            assert_nested_dict(value)
    elif isinstance(obj, list):
        for value in obj:
            assert_nested_dict(value)
    elif isinstance(obj, (int, float, str, bytes, bool, type(None),
                          datetime.datetime)):
        pass
    else:
        assert False, obj


@given(object_dicts())
def test_dicts_generation(obj_type_and_obj):
    (obj_type, object_) = obj_type_and_obj
    assert_nested_dict(object_)
    if obj_type == 'content':
        COMMON_KEYS = set(DEFAULT_ALGORITHMS) | {'length', 'status', 'ctime'}
        if object_['status'] == 'visible':
            assert set(object_) <= COMMON_KEYS | {'data'}
        elif object_['status'] == 'absent':
            assert set(object_) == COMMON_KEYS | {'reason'}
        elif object_['status'] == 'hidden':
            assert set(object_) <= COMMON_KEYS | {'data'}
        else:
            assert False, object_
    elif obj_type == 'release':
        assert object_['target_type'] in target_types
    elif obj_type == 'snapshot':
        for branch in object_['branches'].values():
            assert branch is None or branch['target_type'] in target_types


@given(objects())
def test_model_to_dicts(obj_type_and_obj):
    (obj_type, object_) = obj_type_and_obj
    obj_dict = object_.to_dict()
    assert_nested_dict(obj_dict)
    if obj_type == 'content':
        COMMON_KEYS = set(DEFAULT_ALGORITHMS) | {'length', 'status', 'ctime'}
        if obj_dict['status'] == 'visible':
            assert set(obj_dict) == COMMON_KEYS | {'data'}
        elif obj_dict['status'] == 'absent':
            assert set(obj_dict) == COMMON_KEYS | {'reason'}
        elif obj_dict['status'] == 'hidden':
            assert set(obj_dict) == COMMON_KEYS | {'data'}
        else:
            assert False, obj_dict
    elif obj_type == 'release':
        assert obj_dict['target_type'] in target_types
    elif obj_type == 'snapshot':
        for branch in obj_dict['branches'].values():
            assert branch is None or branch['target_type'] in target_types


_min_snp_size = 10
_max_snp_size = 100


@given(snapshots(min_size=_min_snp_size, max_size=_max_snp_size))
@settings(max_examples=1)
def test_snapshots_strategy(snapshot):

    branches = snapshot.branches

    assert len(branches) >= _min_snp_size
    assert len(branches) <= _max_snp_size

    aliases = []

    # check snapshot integrity
    for name, branch in branches.items():
        assert branch is None or branch.target_type.value in target_types
        if branch is not None and branch.target_type == TargetType.ALIAS:
            aliases.append(name)
            assert branch.target in branches

    # check no cycles between aliases
    for alias in aliases:
        processed_alias = set()
        current_alias = alias
        while (branches[current_alias] is not None
                and branches[current_alias].target_type == TargetType.ALIAS):
            assert branches[current_alias].target not in processed_alias
            processed_alias.add(current_alias)
            current_alias = branches[current_alias].target


@given(snapshots(min_size=_min_snp_size, max_size=_min_snp_size))
@settings(max_examples=1)
def test_snapshots_strategy_fixed_size(snapshot):
    assert len(snapshot.branches) == _min_snp_size



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from hypothesis import assume
from hypothesis.strategies import (
    binary, booleans, builds, characters,
    composite, datetimes, dictionaries, from_regex, integers, just, lists,
    none, one_of, sampled_from, sets, text, )

from .from_disk import DentryPerms
from .model import (
    Person, Timestamp, TimestampWithTimezone, Origin,
    OriginVisit, OriginVisitUpdate, Snapshot, SnapshotBranch, ObjectType,
    TargetType, Release, Revision, RevisionType, BaseContent, Directory,
    DirectoryEntry, Content, SkippedContent, )
from .identifiers import snapshot_identifier, identifier_to_bytes


pgsql_alphabet = characters(
    blacklist_categories=('Cs', ),
    blacklist_characters=['\u0000'])  # postgresql does not like these


def optional(strategy):
    return one_of(none(), strategy)


def pgsql_text():
    return text(alphabet=pgsql_alphabet)


def sha1_git():
    return binary(min_size=20, max_size=20)


def sha1():
    return binary(min_size=20, max_size=20)


@composite
def urls(draw):
    protocol = draw(sampled_from(['git', 'http', 'https', 'deb']))
    domain = draw(from_regex(r'\A([a-z]([a-z0-9-]*)\.){1,3}[a-z0-9]+\Z'))

    return '%s://%s' % (protocol, domain)


def persons_d():
    return builds(
        dict,
        fullname=binary(),
        email=optional(binary()),
        name=optional(binary()),
    )


def persons():
    return persons_d().map(Person.from_dict)


def timestamps_d():
    max_seconds = datetime.datetime.max.replace(
        tzinfo=datetime.timezone.utc).timestamp()
    min_seconds = datetime.datetime.min.replace(
        tzinfo=datetime.timezone.utc).timestamp()
    return builds(
        dict,
        seconds=integers(min_seconds, max_seconds),
        microseconds=integers(0, 1000000))


def timestamps():
    return timestamps_d().map(Timestamp.from_dict)


@composite
def timestamps_with_timezone_d(
        draw,
        timestamp=timestamps_d(),
        offset=integers(min_value=-14*60, max_value=14*60),
        negative_utc=booleans()):
    timestamp = draw(timestamp)
    offset = draw(offset)
    negative_utc = draw(negative_utc)
    assume(not (negative_utc and offset))
    return dict(
        timestamp=timestamp,
        offset=offset,
        negative_utc=negative_utc)


timestamps_with_timezone = timestamps_with_timezone_d().map(
    TimestampWithTimezone.from_dict)


def origins_d():
    return builds(
        dict,
        url=urls())


def origins():
    return origins_d().map(Origin.from_dict)


def origin_visits_d():
    return builds(
        dict,
        visit=integers(0, 1000),
        origin=urls(),
        date=datetimes(),
        status=sampled_from(['ongoing', 'full', 'partial']),
        type=pgsql_text(),
        snapshot=optional(sha1_git()),
    )


def origin_visits():
    return origin_visits_d().map(OriginVisit.from_dict)


def metadata_dicts():
    return dictionaries(pgsql_text(), pgsql_text())


def origin_visit_updates_d():
    return builds(
        dict,
        visit=integers(0, 1000),
        origin=urls(),
        status=sampled_from(['ongoing', 'full', 'partial']),
        date=datetimes(),
        snapshot=optional(sha1_git()),
        metadata=one_of(none(), metadata_dicts()))


def origin_visit_updates():
    return origin_visit_updates_d().map(OriginVisitUpdate.from_dict)


@composite
def releases_d(draw):
    target_type = sampled_from([x.value for x in ObjectType])
    name = binary()
    message = binary()
    synthetic = booleans()
    target = sha1_git()
    metadata = one_of(none(), revision_metadata())

    return draw(one_of(
        builds(
            dict,
            name=name,
            message=message,
            synthetic=synthetic,
            author=none(),
            date=none(),
            target=target,
            target_type=target_type,
            metadata=metadata,
        ),
        builds(
            dict,
            name=name,
            message=message,
            synthetic=synthetic,
            date=timestamps_with_timezone_d(),
            author=persons_d(),
            target=target,
            target_type=target_type,
            metadata=metadata,
        ),
    ))


def releases():
    return releases_d().map(Release.from_dict)


revision_metadata = metadata_dicts


def revisions_d():
    return builds(
        dict,
        message=binary(),
        synthetic=booleans(),
        author=persons_d(),
        committer=persons_d(),
        date=timestamps_with_timezone_d(),
        committer_date=timestamps_with_timezone_d(),
        parents=lists(sha1_git()),
        directory=sha1_git(),
        type=sampled_from([x.value for x in RevisionType]),
        metadata=one_of(none(), revision_metadata()))
    # TODO: metadata['extra_headers'] can have binary keys and values


def revisions():
    return revisions_d().map(Revision.from_dict)


def directory_entries_d():
    return builds(
        dict,
        name=binary(),
        target=sha1_git(),
        type=sampled_from(['file', 'dir', 'rev']),
        perms=sampled_from([perm.value for perm in DentryPerms]))


def directory_entries():
    return directory_entries_d().map(DirectoryEntry)


def directories_d():
    return builds(
        dict,
        entries=lists(directory_entries_d()))


def directories():
    return directories_d().map(Directory.from_dict)


def contents_d():
    return one_of(present_contents_d(), skipped_contents_d())


def contents():
    return one_of(present_contents(), skipped_contents())


def present_contents_d():
    return builds(
        dict,
        data=binary(max_size=4096),
        ctime=optional(datetimes()),
        status=one_of(just('visible'), just('hidden')),
    )


def present_contents():
    return present_contents_d().map(lambda d: Content.from_data(**d))


@composite
def skipped_contents_d(draw):
    result = BaseContent._hash_data(draw(binary(max_size=4096)))
    result.pop('data')
    nullify_attrs = draw(
        sets(sampled_from(['sha1', 'sha1_git', 'sha256', 'blake2s256']))
    )
    for k in nullify_attrs:
        result[k] = None
    result['reason'] = draw(pgsql_text())
    result['status'] = 'absent'
    result['ctime'] = draw(optional(datetimes()))
    return result


def skipped_contents():
    return skipped_contents_d().map(SkippedContent.from_dict)


def branch_names():
    return binary(min_size=1)


def branch_targets_object_d():
    return builds(
        dict,
        target=sha1_git(),
        target_type=sampled_from([
            x.value for x in TargetType
            if x.value not in ('alias', )]))


def branch_targets_alias_d():
    return builds(
        dict,
        target=sha1_git(),
        target_type=just('alias'))  # TargetType.ALIAS.value))


def branch_targets_d(*, only_objects=False):
    if only_objects:
        return branch_targets_object_d()
    else:
        return one_of(branch_targets_alias_d(), branch_targets_object_d())


def branch_targets(*, only_objects=False):
    return builds(
        SnapshotBranch.from_dict,
        branch_targets_d(only_objects=only_objects))


@composite
def snapshots_d(draw, *, min_size=0, max_size=100, only_objects=False):
    branches = draw(dictionaries(
        keys=branch_names(),
        values=one_of(
            none(),
            branch_targets_d(only_objects=only_objects)
        ),
        min_size=min_size,
        max_size=max_size,
    ))

    if not only_objects:
        # Make sure aliases point to actual branches
        unresolved_aliases = {
            branch: target['target']
            for branch, target in branches.items()
            if (target
                and target['target_type'] == 'alias'
                and target['target'] not in branches)
        }
        for alias_name, alias_target in unresolved_aliases.items():
            # Override alias branch with one pointing to a real object
            # if max_size constraint is reached
            alias = alias_target if len(branches) < max_size else alias_name
            branches[alias] = draw(branch_targets_d(only_objects=True))

    # Ensure no cycles between aliases
    while True:
        try:
            id_ = snapshot_identifier({
                'branches': {
                    name: branch or None
                    for (name, branch) in branches.items()}})
        except ValueError as e:
            for (source, target) in e.args[1]:
                branches[source] = draw(branch_targets_d(only_objects=True))
        else:
            break

    return dict(
        id=identifier_to_bytes(id_),
        branches=branches)


def snapshots(*, min_size=0, max_size=100, only_objects=False):
    return snapshots_d(min_size=min_size, max_size=max_size,
                       only_objects=only_objects).map(
        Snapshot.from_dict)


def objects():
    return one_of(
        origins().map(lambda x: ('origin', x)),
        origin_visits().map(lambda x: ('origin_visit', x)),
        origin_visit_updates().map(lambda x: ('origin_visit_update', x)),
        snapshots().map(lambda x: ('snapshot', x)),
        releases().map(lambda x: ('release', x)),
        revisions().map(lambda x: ('revision', x)),
        directories().map(lambda x: ('directory', x)),
        contents().map(lambda x: ('content', x)),
    )


def object_dicts():
    """generates a random couple (type, dict)

    which dict is suitable for <ModelForType>.from_dict() factory methods.
    """
    return one_of(
        origins_d().map(lambda x: ('origin', x)),
        origin_visits_d().map(lambda x: ('origin_visit', x)),
        snapshots_d().map(lambda x: ('snapshot', x)),
        releases_d().map(lambda x: ('release', x)),
        revisions_d().map(lambda x: ('revision', x)),
        directories_d().map(lambda x: ('directory', x)),
        contents_d().map(lambda x: ('content', x)),
    )



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections


def toposort(revision_log):
    """Perform a topological sort on a revision log graph.

    Complexity: O(N) (linear in the length of the revision log)

    Args:
        revision_log: Revision log as returned by
            swh.storage.Storage.revision_log().

    Yields:
        The revision log sorted by a topological order
    """
    in_degree = {}  # rev_id -> numbers of parents left to compute
    children = collections.defaultdict(list)  # rev_id -> children

    # Compute the in_degrees and the parents of all the revisions.
    # Add the roots to the processing queue.
    queue = collections.deque()
    for rev in revision_log:
        parents = rev['parents']
        in_degree[rev['id']] = len(parents)
        if not parents:
            queue.append(rev)
        for parent in parents:
            children[parent].append(rev)

    # Topological sort: yield the 'ready' nodes, decrease the in degree of
    # their children and add the 'ready' ones to the queue.
    while queue:
        rev = queue.popleft()
        yield rev
        for child in children[rev['id']]:
            in_degree[child['id']] -= 1
            if in_degree[child['id']] == 0:
                queue.append(child)



# Copyright (C) 2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from operator import itemgetter
import os
import sys

from swh.model.from_disk import Directory, DentryPerms
from swh.model.hashutil import ALGORITHMS, hash_to_hex


def generate_from_directory(varname, directory, indent=0):
    """Generate test data from a given directory"""
    def get_data(member, path):
        yield (path, member.get_data())
        if isinstance(member, Directory):
            for name, child in member.items():
                yield from get_data(child, os.path.join(path, name))

    data = dict(get_data(directory, b''))
    out = []

    def format_hash(h, indent=0):
        spindent = ' ' * indent
        if len(h) > 20:
            cutoff = len(h)//2
            parts = h[:cutoff], h[cutoff:]
        else:
            parts = [h]

        out.append('hash_to_bytes(\n')
        for part in parts:
            out.append(spindent + '    %s\n' % repr(hash_to_hex(part)))
        out.append(spindent + ')')

    def format_dict_items(d, indent=0):
        spindent = ' ' * indent
        for key, value in sorted(d.items()):
            if isinstance(key, bytes):
                out.append(spindent + repr(key) + ': {\n')
                format_dict_items(value, indent=indent + 4)
                out.append(spindent + '}')
            else:
                out.append(spindent + repr(key) + ': ')
                if key == 'entries':
                    if not value:
                        out.append('[]')
                    else:
                        out.append('[')
                        last_index = len(value) - 1
                        for i, entry in enumerate(
                                sorted(value, key=itemgetter('name'))):
                            if i:
                                out.append(' ')
                            out.append('{\n')
                            format_dict_items(entry, indent=indent + 4)
                            if i != last_index:
                                out.append(spindent + '},')
                        out.append(spindent + '}]')
                elif key in ALGORITHMS | {'id', 'target'}:
                    format_hash(value, indent=indent)
                elif isinstance(value, DentryPerms):
                    out.append(str(value))
                else:
                    out.append(repr(value))
            out.append(',\n')

    spindent = ' ' * indent
    out.append(spindent + '%s = {\n' % varname)
    format_dict_items(data, indent=4 + indent)
    out.append(spindent + '}')

    return ''.join(out)


if __name__ == '__main__':
    if not sys.argv[1:]:
        print("Usage: %s dir1 dir2" % sys.argv[0], file=sys.stderr)
        exit(2)

    for dirname in sys.argv[1:]:
        basename = os.path.basename(dirname)
        varname = 'expected_%s' % basename
        testdata = generate_from_directory(
            varname,
            Directory.from_disk(path=os.fsencode(dirname)),
            indent=8
        )
        print(testdata)
        print()



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import dulwich.repo
import os
import sys

from functools import partial
from urllib.parse import urlparse

from swh.model import hashutil
from swh.model import identifiers as pids
from swh.model.exceptions import ValidationError
from swh.model.from_disk import Content, Directory


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

# Mapping between dulwich types and Software Heritage ones. Used by snapshot ID
# computation.
_DULWICH_TYPES = {
    b'blob': 'content',
    b'tree': 'directory',
    b'commit': 'revision',
    b'tag': 'release',
}


class PidParamType(click.ParamType):
    name = 'persistent identifier'

    def convert(self, value, param, ctx):
        try:
            pids.parse_persistent_identifier(value)
            return value  # return as string, as we need just that
        except ValidationError as e:
            self.fail('%s is not a valid PID. %s.' % (value, e), param, ctx)


def pid_of_file(path):
    object = Content.from_file(path=path).get_data()
    return pids.persistent_identifier(pids.CONTENT, object)


def pid_of_file_content(data):
    object = Content.from_bytes(mode=644, data=data).get_data()
    return pids.persistent_identifier(pids.CONTENT, object)


def pid_of_dir(path):
    object = Directory.from_disk(path=path).get_data()
    return pids.persistent_identifier(pids.DIRECTORY, object)


def pid_of_origin(url):
    pid = pids.PersistentId(object_type='origin',
                            object_id=pids.origin_identifier({'url': url}))
    return str(pid)


def pid_of_git_repo(path):
    repo = dulwich.repo.Repo(path)

    branches = {}
    for ref, target in repo.refs.as_dict().items():
        obj = repo[target]
        if obj:
            branches[ref] = {
                'target': hashutil.bytehex_to_hash(target),
                'target_type': _DULWICH_TYPES[obj.type_name],
            }
        else:
            branches[ref] = None

    for ref, target in repo.refs.get_symrefs().items():
        branches[ref] = {
            'target': target,
            'target_type': 'alias',
        }

    snapshot = {'branches': branches}

    pid = pids.PersistentId(object_type='snapshot',
                            object_id=pids.snapshot_identifier(snapshot))
    return str(pid)


def identify_object(obj_type, follow_symlinks, obj):
    if obj_type == 'auto':
        if obj == '-' or os.path.isfile(obj):
            obj_type = 'content'
        elif os.path.isdir(obj):
            obj_type = 'directory'
        else:
            try:  # URL parsing
                if urlparse(obj).scheme:
                    obj_type = 'origin'
                else:
                    raise ValueError
            except ValueError:
                raise click.BadParameter('cannot detect object type for %s' %
                                         obj)

    pid = None

    if obj == '-':
        content = sys.stdin.buffer.read()
        pid = pid_of_file_content(content)
    elif obj_type in ['content', 'directory']:
        path = obj.encode(sys.getfilesystemencoding())
        if follow_symlinks and os.path.islink(obj):
            path = os.path.realpath(obj)
        if obj_type == 'content':
            pid = pid_of_file(path)
        elif obj_type == 'directory':
            pid = pid_of_dir(path)
    elif obj_type == 'origin':
        pid = pid_of_origin(obj)
    elif obj_type == 'snapshot':
        pid = pid_of_git_repo(obj)
    else:  # shouldn't happen, due to option validation
        raise click.BadParameter('invalid object type: ' + obj_type)

    # note: we return original obj instead of path here, to preserve user-given
    # file name in output
    return (obj, pid)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--dereference/--no-dereference', 'follow_symlinks',
              default=True,
              help='follow (or not) symlinks for OBJECTS passed as arguments '
              + '(default: follow)')
@click.option('--filename/--no-filename', 'show_filename', default=True,
              help='show/hide file name (default: show)')
@click.option('--type', '-t', 'obj_type', default='auto',
              type=click.Choice(['auto', 'content', 'directory', 'origin',
                                 'snapshot']),
              help='type of object to identify (default: auto)')
@click.option('--verify', '-v', metavar='PID', type=PidParamType(),
              help='reference identifier to be compared with computed one')
@click.argument('objects', nargs=-1)
def identify(obj_type, verify, show_filename, follow_symlinks, objects):
    """Compute the Software Heritage persistent identifier (PID) for the given
    source code object(s).

    For more details about Software Heritage PIDs see:

    \b
    https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html

    \b
    Examples:

    \b
      $ swh identify fork.c kmod.c sched/deadline.c
      swh:1:cnt:2e391c754ae730bd2d8520c2ab497c403220c6e3    fork.c
      swh:1:cnt:0277d1216f80ae1adeed84a686ed34c9b2931fc2    kmod.c
      swh:1:cnt:57b939c81bce5d06fa587df8915f05affbe22b82    sched/deadline.c

    \b
      $ swh identify --no-filename /usr/src/linux/kernel/
      swh:1:dir:f9f858a48d663b3809c9e2f336412717496202ab

    \b
      $ git clone --mirror https://forge.softwareheritage.org/source/helloworld.git
      $ swh identify --type snapshot helloworld.git/
      swh:1:snp:510aa88bdc517345d258c1fc2babcd0e1f905e93	helloworld.git

    """  # NoQA  # overlong lines in shell examples are fine
    if not objects:
        objects = ['-']

    if verify and len(objects) != 1:
        raise click.BadParameter('verification requires a single object')

    results = map(partial(identify_object, obj_type, follow_symlinks), objects)

    if verify:
        pid = next(results)[1]
        if verify == pid:
            click.echo('PID match: %s' % pid)
            sys.exit(0)
        else:
            click.echo('PID mismatch: %s != %s' % (verify, pid))
            sys.exit(1)
    else:
        for (obj, pid) in results:
            msg = pid
            if show_filename:
                msg = '%s\t%s' % (pid, os.fsdecode(obj))
            click.echo(msg)


if __name__ == '__main__':
    identify()



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from .exceptions import ValidationError, NON_FIELD_ERRORS
from . import fields
from .hashutil import MultiHash, hash_to_bytes


def validate_content(content):
    """Validate that a content has the correct schema.

    Args: a content (dictionary) to validate."""

    def validate_content_status(status):
        return fields.validate_enum(status, {'absent', 'visible', 'hidden'})

    def validate_keys(content):
        hashes = {'sha1', 'sha1_git', 'sha256'}
        errors = []

        out = True
        if content['status'] == 'absent':
            try:
                out = out and fields.validate_all_keys(content, {'reason',
                                                                 'origin'})
            except ValidationError as e:
                errors.append(e)
            try:
                out = out and fields.validate_any_key(content, hashes)
            except ValidationError as e:
                errors.append(e)
        else:
            try:
                out = out and fields.validate_all_keys(content, hashes)
            except ValidationError as e:
                errors.append(e)

        if errors:
            raise ValidationError(errors)

        return out

    def validate_hashes(content):
        errors = []
        if 'data' in content:
            hashes = MultiHash.from_data(content['data']).digest()
            for hash_type, computed_hash in hashes.items():
                if hash_type not in content:
                    continue
                content_hash = hash_to_bytes(content[hash_type])
                if content_hash != computed_hash:
                    errors.append(ValidationError(
                        'hash mismatch in content for hash %(hash)s',
                        params={'hash': hash_type},
                        code='content-hash-mismatch',
                    ))
            if errors:
                raise ValidationError(errors)

        return True

    content_schema = {
        'sha1': (False, fields.validate_sha1),
        'sha1_git': (False, fields.validate_sha1_git),
        'sha256': (False, fields.validate_sha256),
        'status': (True, validate_content_status),
        'length': (True, fields.validate_int),
        'ctime': (True, fields.validate_datetime),
        'reason': (False, fields.validate_str),
        'origin': (False, fields.validate_int),
        'data': (False, fields.validate_bytes),
        NON_FIELD_ERRORS: [validate_keys, validate_hashes],
    }

    return fields.validate_against_schema('content', content_schema, content)



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import unittest

from swh.model.exceptions import ValidationError
from swh.model.fields import simple


class ValidateSimple(unittest.TestCase):
    def setUp(self):
        self.valid_str = 'I am a valid string'

        self.valid_bytes = b'I am a valid bytes object'

        self.enum_values = {'an enum value', 'other', 'and another'}
        self.invalid_enum_value = 'invalid enum value'

        self.valid_int = 42

        self.valid_real = 42.42

        self.valid_datetime = datetime.datetime(1999, 1, 1, 12, 0, 0,
                                                tzinfo=datetime.timezone.utc)
        self.invalid_datetime_notz = datetime.datetime(1999, 1, 1, 12, 0, 0)

    def test_validate_int(self):
        self.assertTrue(simple.validate_int(self.valid_int))

    def test_validate_int_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_int(self.valid_str)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-type')
        self.assertEqual(exc.params['expected_type'], 'Integral')
        self.assertEqual(exc.params['type'], 'str')

    def test_validate_str(self):
        self.assertTrue(simple.validate_str(self.valid_str))

    def test_validate_str_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_str(self.valid_int)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-type')
        self.assertEqual(exc.params['expected_type'], 'str')
        self.assertEqual(exc.params['type'], 'int')

        with self.assertRaises(ValidationError) as cm:
            simple.validate_str(self.valid_bytes)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-type')
        self.assertEqual(exc.params['expected_type'], 'str')
        self.assertEqual(exc.params['type'], 'bytes')

    def test_validate_bytes(self):
        self.assertTrue(simple.validate_bytes(self.valid_bytes))

    def test_validate_bytes_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_bytes(self.valid_int)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-type')
        self.assertEqual(exc.params['expected_type'], 'bytes')
        self.assertEqual(exc.params['type'], 'int')

        with self.assertRaises(ValidationError) as cm:
            simple.validate_bytes(self.valid_str)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-type')
        self.assertEqual(exc.params['expected_type'], 'bytes')
        self.assertEqual(exc.params['type'], 'str')

    def test_validate_datetime(self):
        self.assertTrue(simple.validate_datetime(self.valid_datetime))
        self.assertTrue(simple.validate_datetime(self.valid_int))
        self.assertTrue(simple.validate_datetime(self.valid_real))

    def test_validate_datetime_invalid_type(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_datetime(self.valid_str)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-type')
        self.assertEqual(exc.params['expected_type'], 'one of datetime, Real')
        self.assertEqual(exc.params['type'], 'str')

    def test_validate_datetime_invalide_tz(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_datetime(self.invalid_datetime_notz)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'datetime-without-tzinfo')

    def test_validate_enum(self):
        for value in self.enum_values:
            self.assertTrue(simple.validate_enum(value, self.enum_values))

    def test_validate_enum_invalid_value(self):
        with self.assertRaises(ValidationError) as cm:
            simple.validate_enum(self.invalid_enum_value, self.enum_values)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, 'unexpected-value')
        self.assertEqual(exc.params['value'], self.invalid_enum_value)
        self.assertEqual(exc.params['expected_values'],
                         ', '.join(sorted(self.enum_values)))



# Copyright (C) 2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import pytest
import tarfile
import tempfile
import unittest

from typing import ClassVar, Optional

from swh.model import from_disk
from swh.model.from_disk import (
    Content, DentryPerms, Directory, DiskBackedContent
)
from swh.model.hashutil import DEFAULT_ALGORITHMS, hash_to_bytes, hash_to_hex
from swh.model import model

TEST_DATA = os.path.join(os.path.dirname(__file__), 'data')


class ModeToPerms(unittest.TestCase):
    def setUp(self):
        super().setUp()

        # Generate a full permissions map
        self.perms_map = {}

        # Symlinks
        for i in range(0o120000, 0o127777 + 1):
            self.perms_map[i] = DentryPerms.symlink

        # Directories
        for i in range(0o040000, 0o047777 + 1):
            self.perms_map[i] = DentryPerms.directory

        # Other file types: socket, regular file, block device, character
        # device, fifo all map to regular files
        for ft in [0o140000, 0o100000, 0o060000, 0o020000, 0o010000]:
            for i in range(ft, ft + 0o7777 + 1):
                if i & 0o111:
                    # executable bits are set
                    self.perms_map[i] = DentryPerms.executable_content
                else:
                    self.perms_map[i] = DentryPerms.content

    def test_exhaustive_mode_to_perms(self):
        for fmode, perm in self.perms_map.items():
            self.assertEqual(perm, from_disk.mode_to_perms(fmode))


class TestDiskBackedContent(unittest.TestCase):
    def test_with_data(self):
        expected_content = model.Content(
            length=42, status='visible', data=b'foo bar',
            sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')
        with tempfile.NamedTemporaryFile(mode='w+b') as fd:
            content = DiskBackedContent(
                length=42, status='visible', path=fd.name,
                sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')
            fd.write(b'foo bar')
            fd.seek(0)
            content_with_data = content.with_data()

        assert expected_content == content_with_data

    def test_lazy_data(self):
        with tempfile.NamedTemporaryFile(mode='w+b') as fd:
            fd.write(b'foo')
            fd.seek(0)
            content = DiskBackedContent(
                length=42, status='visible', path=fd.name,
                sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')
            fd.write(b'bar')
            fd.seek(0)
            content_with_data = content.with_data()
            fd.write(b'baz')
            fd.seek(0)

        assert content_with_data.data == b'bar'

    def test_with_data_cannot_read(self):
        with tempfile.NamedTemporaryFile(mode='w+b') as fd:
            content = DiskBackedContent(
                length=42, status='visible', path=fd.name,
                sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')

        with pytest.raises(OSError):
            content.with_data()

    def test_missing_path(self):
        with pytest.raises(TypeError):
            DiskBackedContent(
                length=42, status='visible',
                sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')

        with pytest.raises(TypeError):
            DiskBackedContent(
                length=42, status='visible', path=None,
                sha1=b'foo', sha1_git=b'bar', sha256=b'baz', blake2s256=b'qux')


class DataMixin:
    maxDiff = None  # type: ClassVar[Optional[int]]

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(
            prefix='swh.model.from_disk'
        )
        self.tmpdir_name = os.fsencode(self.tmpdir.name)

        self.contents = {
            b'file': {
                'data': b'42\n',
                'sha1': hash_to_bytes(
                    '34973274ccef6ab4dfaaf86599792fa9c3fe4689'
                ),
                'sha256': hash_to_bytes(
                    '084c799cd551dd1d8d5c5f9a5d593b2e'
                    '931f5e36122ee5c793c1d08a19839cc0'
                ),
                'sha1_git': hash_to_bytes(
                    'd81cc0710eb6cf9efd5b920a8453e1e07157b6cd'),
                'blake2s256': hash_to_bytes(
                    'd5fe1939576527e42cfd76a9455a2432'
                    'fe7f56669564577dd93c4280e76d661d'
                ),
                'length': 3,
                'mode': 0o100644
            },
        }

        self.symlinks = {
            b'symlink': {
                'data': b'target',
                'blake2s256': hash_to_bytes(
                    '595d221b30fdd8e10e2fdf18376e688e'
                    '9f18d56fd9b6d1eb6a822f8c146c6da6'
                ),
                'sha1': hash_to_bytes(
                    '0e8a3ad980ec179856012b7eecf4327e99cd44cd'
                ),
                'sha1_git': hash_to_bytes(
                    '1de565933b05f74c75ff9a6520af5f9f8a5a2f1d'
                ),
                'sha256': hash_to_bytes(
                    '34a04005bcaf206eec990bd9637d9fdb'
                    '6725e0a0c0d4aebf003f17f4c956eb5c'
                ),
                'length': 6,
                'perms': DentryPerms.symlink,
            }
        }

        self.specials = {
            b'fifo': os.mkfifo,
        }

        self.empty_content = {
            'data': b'',
            'length': 0,
            'blake2s256': hash_to_bytes(
                '69217a3079908094e11121d042354a7c'
                '1f55b6482ca1a51e1b250dfd1ed0eef9'
            ),
            'sha1': hash_to_bytes(
                'da39a3ee5e6b4b0d3255bfef95601890afd80709'
            ),
            'sha1_git': hash_to_bytes(
                'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391'
            ),
            'sha256': hash_to_bytes(
                'e3b0c44298fc1c149afbf4c8996fb924'
                '27ae41e4649b934ca495991b7852b855'
            ),
            'perms': DentryPerms.content,
        }

        self.empty_directory = {
            'id': hash_to_bytes(
                '4b825dc642cb6eb9a060e54bf8d69288fbee4904'
            ),
            'entries': [],
        }

        # Generated with generate_testdata_from_disk
        self.tarball_contents = {
            b'': {
                'entries': [{
                    'name': b'bar',
                    'perms': DentryPerms.directory,
                    'target': hash_to_bytes(
                        '3c1f578394f4623f74a0ba7fe761729f59fc6ec4'
                    ),
                    'type': 'dir',
                }, {
                    'name': b'empty-folder',
                    'perms': DentryPerms.directory,
                    'target': hash_to_bytes(
                        '4b825dc642cb6eb9a060e54bf8d69288fbee4904'
                    ),
                    'type': 'dir',
                }, {
                    'name': b'foo',
                    'perms': DentryPerms.directory,
                    'target': hash_to_bytes(
                        '2b41c40f0d1fbffcba12497db71fba83fcca96e5'
                    ),
                    'type': 'dir',
                }, {
                    'name': b'link-to-another-quote',
                    'perms': DentryPerms.symlink,
                    'target': hash_to_bytes(
                        '7d5c08111e21c8a9f71540939998551683375fad'
                    ),
                    'type': 'file',
                }, {
                    'name': b'link-to-binary',
                    'perms': DentryPerms.symlink,
                    'target': hash_to_bytes(
                        'e86b45e538d9b6888c969c89fbd22a85aa0e0366'
                    ),
                    'type': 'file',
                }, {
                    'name': b'link-to-foo',
                    'perms': DentryPerms.symlink,
                    'target': hash_to_bytes(
                        '19102815663d23f8b75a47e7a01965dcdc96468c'
                    ),
                    'type': 'file',
                }, {
                    'name': b'some-binary',
                    'perms': DentryPerms.executable_content,
                    'target': hash_to_bytes(
                        '68769579c3eaadbe555379b9c3538e6628bae1eb'
                    ),
                    'type': 'file',
                }],
                'id': hash_to_bytes(
                    'e8b0f1466af8608c8a3fb9879db172b887e80759'
                ),
            },
            b'bar': {
                'entries': [{
                    'name': b'barfoo',
                    'perms': DentryPerms.directory,
                    'target': hash_to_bytes(
                        'c3020f6bf135a38c6df3afeb5fb38232c5e07087'
                    ),
                    'type': 'dir',
                }],
                'id': hash_to_bytes(
                    '3c1f578394f4623f74a0ba7fe761729f59fc6ec4'
                ),
            },
            b'bar/barfoo': {
                'entries': [{
                    'name': b'another-quote.org',
                    'perms': DentryPerms.content,
                    'target': hash_to_bytes(
                        '133693b125bad2b4ac318535b84901ebb1f6b638'
                    ),
                    'type': 'file',
                }],
                'id': hash_to_bytes(
                    'c3020f6bf135a38c6df3afeb5fb38232c5e07087'
                ),
            },
            b'bar/barfoo/another-quote.org': {
                'blake2s256': hash_to_bytes(
                    'd26c1cad82d43df0bffa5e7be11a60e3'
                    '4adb85a218b433cbce5278b10b954fe8'
                ),
                'length': 72,
                'perms': DentryPerms.content,
                'sha1': hash_to_bytes(
                    '90a6138ba59915261e179948386aa1cc2aa9220a'
                ),
                'sha1_git': hash_to_bytes(
                    '133693b125bad2b4ac318535b84901ebb1f6b638'
                ),
                'sha256': hash_to_bytes(
                    '3db5ae168055bcd93a4d08285dc99ffe'
                    'e2883303b23fac5eab850273a8ea5546'
                ),
            },
            b'empty-folder': {
                'entries': [],
                'id': hash_to_bytes(
                    '4b825dc642cb6eb9a060e54bf8d69288fbee4904'
                ),
            },
            b'foo': {
                'entries': [{
                    'name': b'barfoo',
                    'perms': DentryPerms.symlink,
                    'target': hash_to_bytes(
                        '8185dfb2c0c2c597d16f75a8a0c37668567c3d7e'
                    ),
                    'type': 'file',
                }, {
                    'name': b'quotes.md',
                    'perms': DentryPerms.content,
                    'target': hash_to_bytes(
                        '7c4c57ba9ff496ad179b8f65b1d286edbda34c9a'
                    ),
                    'type': 'file',
                }, {
                    'name': b'rel-link-to-barfoo',
                    'perms': DentryPerms.symlink,
                    'target': hash_to_bytes(
                        'acac326ddd63b0bc70840659d4ac43619484e69f'
                    ),
                    'type': 'file',
                }],
                'id': hash_to_bytes(
                    '2b41c40f0d1fbffcba12497db71fba83fcca96e5'
                ),
            },
            b'foo/barfoo': {
                'blake2s256': hash_to_bytes(
                    'e1252f2caa4a72653c4efd9af871b62b'
                    'f2abb7bb2f1b0e95969204bd8a70d4cd'
                ),
                'data': b'bar/barfoo',
                'length': 10,
                'perms': DentryPerms.symlink,
                'sha1': hash_to_bytes(
                    '9057ee6d0162506e01c4d9d5459a7add1fedac37'
                ),
                'sha1_git': hash_to_bytes(
                    '8185dfb2c0c2c597d16f75a8a0c37668567c3d7e'
                ),
                'sha256': hash_to_bytes(
                    '29ad3f5725321b940332c78e403601af'
                    'ff61daea85e9c80b4a7063b6887ead68'
                ),
            },
            b'foo/quotes.md': {
                'blake2s256': hash_to_bytes(
                    'bf7ce4fe304378651ee6348d3e9336ed'
                    '5ad603d33e83c83ba4e14b46f9b8a80b'
                ),
                'length': 66,
                'perms': DentryPerms.content,
                'sha1': hash_to_bytes(
                    '1bf0bb721ac92c18a19b13c0eb3d741cbfadebfc'
                ),
                'sha1_git': hash_to_bytes(
                    '7c4c57ba9ff496ad179b8f65b1d286edbda34c9a'
                ),
                'sha256': hash_to_bytes(
                    'caca942aeda7b308859eb56f909ec96d'
                    '07a499491690c453f73b9800a93b1659'
                ),
            },
            b'foo/rel-link-to-barfoo': {
                'blake2s256': hash_to_bytes(
                    'd9c327421588a1cf61f316615005a2e9'
                    'c13ac3a4e96d43a24138d718fa0e30db'
                ),
                'data': b'../bar/barfoo',
                'length': 13,
                'perms': DentryPerms.symlink,
                'sha1': hash_to_bytes(
                    'dc51221d308f3aeb2754db48391b85687c2869f4'
                ),
                'sha1_git': hash_to_bytes(
                    'acac326ddd63b0bc70840659d4ac43619484e69f'
                ),
                'sha256': hash_to_bytes(
                    '8007d20db2af40435f42ddef4b8ad76b'
                    '80adbec26b249fdf0473353f8d99df08'
                ),
            },
            b'link-to-another-quote': {
                'blake2s256': hash_to_bytes(
                    '2d0e73cea01ba949c1022dc10c8a43e6'
                    '6180639662e5dc2737b843382f7b1910'
                ),
                'data': b'bar/barfoo/another-quote.org',
                'length': 28,
                'perms': DentryPerms.symlink,
                'sha1': hash_to_bytes(
                    'cbeed15e79599c90de7383f420fed7acb48ea171'
                ),
                'sha1_git': hash_to_bytes(
                    '7d5c08111e21c8a9f71540939998551683375fad'
                ),
                'sha256': hash_to_bytes(
                    'e6e17d0793aa750a0440eb9ad5b80b25'
                    '8076637ef0fb68f3ac2e59e4b9ac3ba6'
                ),
            },
            b'link-to-binary': {
                'blake2s256': hash_to_bytes(
                    '9ce18b1adecb33f891ca36664da676e1'
                    '2c772cc193778aac9a137b8dc5834b9b'
                ),
                'data': b'some-binary',
                'length': 11,
                'perms': DentryPerms.symlink,
                'sha1': hash_to_bytes(
                    'd0248714948b3a48a25438232a6f99f0318f59f1'
                ),
                'sha1_git': hash_to_bytes(
                    'e86b45e538d9b6888c969c89fbd22a85aa0e0366'
                ),
                'sha256': hash_to_bytes(
                    '14126e97d83f7d261c5a6889cee73619'
                    '770ff09e40c5498685aba745be882eff'
                ),
            },
            b'link-to-foo': {
                'blake2s256': hash_to_bytes(
                    '08d6cad88075de8f192db097573d0e82'
                    '9411cd91eb6ec65e8fc16c017edfdb74'
                ),
                'data': b'foo',
                'length': 3,
                'perms': DentryPerms.symlink,
                'sha1': hash_to_bytes(
                    '0beec7b5ea3f0fdbc95d0dd47f3c5bc275da8a33'
                ),
                'sha1_git': hash_to_bytes(
                    '19102815663d23f8b75a47e7a01965dcdc96468c'
                ),
                'sha256': hash_to_bytes(
                    '2c26b46b68ffc68ff99b453c1d304134'
                    '13422d706483bfa0f98a5e886266e7ae'
                ),
            },
            b'some-binary': {
                'blake2s256': hash_to_bytes(
                    '922e0f7015035212495b090c27577357'
                    'a740ddd77b0b9e0cd23b5480c07a18c6'
                ),
                'length': 5,
                'perms': DentryPerms.executable_content,
                'sha1': hash_to_bytes(
                    '0bbc12d7f4a2a15b143da84617d95cb223c9b23c'
                ),
                'sha1_git': hash_to_bytes(
                    '68769579c3eaadbe555379b9c3538e6628bae1eb'
                ),
                'sha256': hash_to_bytes(
                    'bac650d34a7638bb0aeb5342646d24e3'
                    'b9ad6b44c9b383621faa482b990a367d'
                ),
            },
        }

    def tearDown(self):
        self.tmpdir.cleanup()

    def assertContentEqual(self, left, right, *,  # noqa
                           check_path=False):
        if not isinstance(left, Content):
            raise ValueError('%s is not a Content' % left)
        if isinstance(right, Content):
            right = right.get_data()

        # Compare dictionaries

        keys = DEFAULT_ALGORITHMS | {
            'length',
            'perms',
        }
        if check_path:
            keys |= {'path'}

        failed = []
        for key in keys:
            try:
                lvalue = left.data[key]
                if key == 'perms' and 'perms' not in right:
                    rvalue = from_disk.mode_to_perms(right['mode'])
                else:
                    rvalue = right[key]
            except KeyError:
                failed.append(key)
                continue

            if lvalue != rvalue:
                failed.append(key)

        if failed:
            raise self.failureException(
                'Content mismatched:\n' +
                '\n'.join(
                    'content[%s] = %r != %r' % (
                        key, left.data.get(key), right.get(key))
                    for key in failed
                )
            )

    def assertDirectoryEqual(self, left, right):  # NoQA
        if not isinstance(left, Directory):
            raise ValueError('%s is not a Directory' % left)
        if isinstance(right, Directory):
            right = right.get_data()

        assert left.entries == right['entries']
        assert left.hash == right['id']

        assert left.to_model() == model.Directory.from_dict(right)

    def make_contents(self, directory):
        for filename, content in self.contents.items():
            path = os.path.join(directory, filename)
            with open(path, 'wb') as f:
                f.write(content['data'])
            os.chmod(path, content['mode'])

    def make_symlinks(self, directory):
        for filename, symlink in self.symlinks.items():
            path = os.path.join(directory, filename)
            os.symlink(symlink['data'], path)

    def make_specials(self, directory):
        for filename, fn in self.specials.items():
            path = os.path.join(directory, filename)
            fn(path)

    def make_from_tarball(self, directory):
        tarball = os.path.join(TEST_DATA, 'dir-folders', 'sample-folder.tgz')

        with tarfile.open(tarball, 'r:gz') as f:
            f.extractall(os.fsdecode(directory))


class TestContent(DataMixin, unittest.TestCase):
    def setUp(self):
        super().setUp()

    def test_data_to_content(self):
        for filename, content in self.contents.items():
            conv_content = Content.from_bytes(mode=content['mode'],
                                              data=content['data'])
            self.assertContentEqual(conv_content, content)
            self.assertIn(hash_to_hex(conv_content.hash), repr(conv_content))


class SymlinkToContent(DataMixin, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.make_symlinks(self.tmpdir_name)

    def test_symlink_to_content(self):
        for filename, symlink in self.symlinks.items():
            path = os.path.join(self.tmpdir_name, filename)
            perms = 0o120000
            conv_content = Content.from_symlink(path=path, mode=perms)
            self.assertContentEqual(conv_content, symlink)

    def test_symlink_to_base_model(self):
        for filename, symlink in self.symlinks.items():
            path = os.path.join(self.tmpdir_name, filename)
            perms = 0o120000
            model_content = \
                Content.from_symlink(path=path, mode=perms).to_model()

            right = symlink.copy()
            for key in ('perms', 'path', 'mode'):
                right.pop(key, None)
            right['status'] = 'visible'
            assert model_content == model.Content.from_dict(right)


class FileToContent(DataMixin, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.make_contents(self.tmpdir_name)
        self.make_symlinks(self.tmpdir_name)
        self.make_specials(self.tmpdir_name)

    def test_symlink_to_content(self):
        for filename, symlink in self.symlinks.items():
            path = os.path.join(self.tmpdir_name, filename)
            conv_content = Content.from_file(path=path)
            self.assertContentEqual(conv_content, symlink)

    def test_file_to_content(self):
        for filename, content in self.contents.items():
            path = os.path.join(self.tmpdir_name, filename)
            conv_content = Content.from_file(path=path)
            self.assertContentEqual(conv_content, content)

    def test_special_to_content(self):
        for filename in self.specials:
            path = os.path.join(self.tmpdir_name, filename)
            conv_content = Content.from_file(path=path)
            self.assertContentEqual(conv_content, self.empty_content)

        for path in ['/dev/null', '/dev/zero']:
            path = os.path.join(self.tmpdir_name, filename)
            conv_content = Content.from_file(path=path)
            self.assertContentEqual(conv_content, self.empty_content)

    def test_symlink_to_content_model(self):
        for filename, symlink in self.symlinks.items():
            path = os.path.join(self.tmpdir_name, filename)
            model_content = Content.from_file(path=path).to_model()

            right = symlink.copy()
            for key in ('perms', 'path', 'mode'):
                right.pop(key, None)
            right['status'] = 'visible'
            assert model_content == model.Content.from_dict(right)

    def test_file_to_content_model(self):
        for filename, content in self.contents.items():
            path = os.path.join(self.tmpdir_name, filename)
            model_content = Content.from_file(path=path).to_model()

            right = content.copy()
            for key in ('perms', 'mode'):
                right.pop(key, None)
            assert model_content.with_data() == model.Content.from_dict(right)

            right['path'] = path
            del right['data']
            assert model_content == DiskBackedContent.from_dict(right)

    def test_special_to_content_model(self):
        for filename in self.specials:
            path = os.path.join(self.tmpdir_name, filename)
            model_content = Content.from_file(path=path).to_model()

            right = self.empty_content.copy()
            for key in ('perms', 'path', 'mode'):
                right.pop(key, None)
            right['status'] = 'visible'
            assert model_content == model.Content.from_dict(right)

        for path in ['/dev/null', '/dev/zero']:
            model_content = Content.from_file(path=path).to_model()

            right = self.empty_content.copy()
            for key in ('perms', 'path', 'mode'):
                right.pop(key, None)
            right['status'] = 'visible'
            assert model_content == model.Content.from_dict(right)

    def test_symlink_max_length(self):
        for max_content_length in [4, 10]:
            for filename, symlink in self.symlinks.items():
                path = os.path.join(self.tmpdir_name, filename)
                content = Content.from_file(path=path)
                if content.data['length'] > max_content_length:
                    with pytest.raises(Exception, match='too large'):
                        Content.from_file(
                            path=path,
                            max_content_length=max_content_length)
                else:
                    limited_content = Content.from_file(
                        path=path,
                        max_content_length=max_content_length)
                    assert content == limited_content

    def test_file_max_length(self):
        for max_content_length in [2, 4]:
            for filename, content in self.contents.items():
                path = os.path.join(self.tmpdir_name, filename)
                content = Content.from_file(path=path)
                limited_content = Content.from_file(
                    path=path,
                    max_content_length=max_content_length)
                assert content.data['length'] == limited_content.data['length']
                assert content.data['status'] == 'visible'
                if content.data['length'] > max_content_length:
                    assert limited_content.data['status'] == 'absent'
                    assert limited_content.data['reason'] \
                        == 'Content too large'
                else:
                    assert limited_content.data['status'] == 'visible'

    def test_special_file_max_length(self):
        for max_content_length in [None, 0, 1]:
            for filename in self.specials:
                path = os.path.join(self.tmpdir_name, filename)
                content = Content.from_file(path=path)
                limited_content = Content.from_file(
                    path=path,
                    max_content_length=max_content_length)
                assert limited_content == content

    def test_file_to_content_with_path(self):
        for filename, content in self.contents.items():
            content_w_path = content.copy()
            path = os.path.join(self.tmpdir_name, filename)
            content_w_path['path'] = path
            conv_content = Content.from_file(path=path)
            self.assertContentEqual(conv_content, content_w_path,
                                    check_path=True)


@pytest.mark.fs
class DirectoryToObjects(DataMixin, unittest.TestCase):
    def setUp(self):
        super().setUp()
        contents = os.path.join(self.tmpdir_name, b'contents')
        os.mkdir(contents)
        self.make_contents(contents)
        symlinks = os.path.join(self.tmpdir_name, b'symlinks')
        os.mkdir(symlinks)
        self.make_symlinks(symlinks)
        specials = os.path.join(self.tmpdir_name, b'specials')
        os.mkdir(specials)
        self.make_specials(specials)
        empties = os.path.join(self.tmpdir_name, b'empty1', b'empty2')
        os.makedirs(empties)

    def test_directory_to_objects(self):
        directory = Directory.from_disk(path=self.tmpdir_name)

        for name, value in self.contents.items():
            self.assertContentEqual(directory[b'contents/' + name], value)

        for name, value in self.symlinks.items():
            self.assertContentEqual(directory[b'symlinks/' + name], value)

        for name in self.specials:
            self.assertContentEqual(
                directory[b'specials/' + name],
                self.empty_content,
            )

        self.assertEqual(
            directory[b'empty1/empty2'].get_data(),
            self.empty_directory,
        )

        # Raise on non existent file
        with self.assertRaisesRegex(KeyError, "b'nonexistent'"):
            directory[b'empty1/nonexistent']

        # Raise on non existent directory
        with self.assertRaisesRegex(KeyError, "b'nonexistentdir'"):
            directory[b'nonexistentdir/file']

        objs = directory.collect()

        self.assertCountEqual(['content', 'directory'], objs)

        self.assertEqual(len(objs['directory']), 6)
        self.assertEqual(len(objs['content']),
                         len(self.contents)
                         + len(self.symlinks)
                         + 1)

    def test_directory_to_objects_ignore_empty(self):
        directory = Directory.from_disk(
            path=self.tmpdir_name,
            dir_filter=from_disk.ignore_empty_directories
        )

        for name, value in self.contents.items():
            self.assertContentEqual(directory[b'contents/' + name], value)

        for name, value in self.symlinks.items():
            self.assertContentEqual(directory[b'symlinks/' + name], value)

        for name in self.specials:
            self.assertContentEqual(
                directory[b'specials/' + name],
                self.empty_content,
            )

        # empty directories have been ignored recursively
        with self.assertRaisesRegex(KeyError, "b'empty1'"):
            directory[b'empty1']
        with self.assertRaisesRegex(KeyError, "b'empty1'"):
            directory[b'empty1/empty2']

        objs = directory.collect()

        self.assertCountEqual(['content', 'directory'], objs)

        self.assertEqual(len(objs['directory']), 4)
        self.assertEqual(len(objs['content']),
                         len(self.contents)
                         + len(self.symlinks)
                         + 1)

    def test_directory_to_objects_ignore_name(self):
        directory = Directory.from_disk(
            path=self.tmpdir_name,
            dir_filter=from_disk.ignore_named_directories([b'symlinks'])
        )
        for name, value in self.contents.items():
            self.assertContentEqual(directory[b'contents/' + name], value)

        for name in self.specials:
            self.assertContentEqual(
                directory[b'specials/' + name],
                self.empty_content,
            )

        self.assertEqual(
            directory[b'empty1/empty2'].get_data(),
            self.empty_directory,
        )

        with self.assertRaisesRegex(KeyError, "b'symlinks'"):
            directory[b'symlinks']

        objs = directory.collect()

        self.assertCountEqual(['content', 'directory'], objs)

        self.assertEqual(len(objs['directory']), 5)
        self.assertEqual(len(objs['content']),
                         len(self.contents)
                         + 1)

    def test_directory_to_objects_ignore_name_case(self):
        directory = Directory.from_disk(
            path=self.tmpdir_name,
            dir_filter=from_disk.ignore_named_directories([b'symLiNks'],
                                                          case_sensitive=False)
        )
        for name, value in self.contents.items():
            self.assertContentEqual(directory[b'contents/' + name], value)

        for name in self.specials:
            self.assertContentEqual(
                directory[b'specials/' + name],
                self.empty_content,
            )

        self.assertEqual(
            directory[b'empty1/empty2'].get_data(),
            self.empty_directory,
        )

        with self.assertRaisesRegex(KeyError, "b'symlinks'"):
            directory[b'symlinks']

        objs = directory.collect()

        self.assertCountEqual(['content', 'directory'], objs)

        self.assertEqual(len(objs['directory']), 5)
        self.assertEqual(len(objs['content']),
                         len(self.contents)
                         + 1)

    def test_directory_entry_order(self):
        with tempfile.TemporaryDirectory() as dirname:
            dirname = os.fsencode(dirname)
            open(os.path.join(dirname, b'foo.'), 'a')
            open(os.path.join(dirname, b'foo0'), 'a')
            os.mkdir(os.path.join(dirname, b'foo'))

            directory = Directory.from_disk(path=dirname)

        assert [entry['name'] for entry in directory.entries] \
            == [b'foo.', b'foo', b'foo0']


@pytest.mark.fs
class TarballTest(DataMixin, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.make_from_tarball(self.tmpdir_name)

    def test_contents_match(self):
        directory = Directory.from_disk(
            path=os.path.join(self.tmpdir_name, b'sample-folder')
        )

        for name, expected in self.tarball_contents.items():
            obj = directory[name]
            if isinstance(obj, Content):
                self.assertContentEqual(obj, expected)
            elif isinstance(obj, Directory):
                self.assertDirectoryEqual(obj, expected)
            else:
                raise self.failureException('Unknown type for %s' % obj)


class DirectoryManipulation(DataMixin, unittest.TestCase):
    def test_directory_access_nested(self):
        d = Directory()
        d[b'a'] = Directory()
        d[b'a/b'] = Directory()

        self.assertEqual(d[b'a/b'].get_data(), self.empty_directory)

    def test_directory_del_nested(self):
        d = Directory()
        d[b'a'] = Directory()
        d[b'a/b'] = Directory()

        with self.assertRaisesRegex(KeyError, "b'c'"):
            del d[b'a/b/c']

        with self.assertRaisesRegex(KeyError, "b'level2'"):
            del d[b'a/level2/c']

        del d[b'a/b']

        self.assertEqual(d[b'a'].get_data(), self.empty_directory)

    def test_directory_access_self(self):
        d = Directory()
        self.assertIs(d, d[b''])
        self.assertIs(d, d[b'/'])
        self.assertIs(d, d[b'//'])

    def test_directory_access_wrong_type(self):
        d = Directory()
        with self.assertRaisesRegex(ValueError, 'bytes from Directory'):
            d['foo']
        with self.assertRaisesRegex(ValueError, 'bytes from Directory'):
            d[42]

    def test_directory_repr(self):
        entries = [b'a', b'b', b'c']
        d = Directory()
        for entry in entries:
            d[entry] = Directory()

        r = repr(d)
        self.assertIn(hash_to_hex(d.hash), r)

        for entry in entries:
            self.assertIn(str(entry), r)

    def test_directory_set_wrong_type_name(self):
        d = Directory()
        with self.assertRaisesRegex(ValueError, 'bytes Directory entry'):
            d['foo'] = Directory()
        with self.assertRaisesRegex(ValueError, 'bytes Directory entry'):
            d[42] = Directory()

    def test_directory_set_nul_in_name(self):
        d = Directory()

        with self.assertRaisesRegex(ValueError, 'nul bytes'):
            d[b'\x00\x01'] = Directory()

    def test_directory_set_empty_name(self):
        d = Directory()
        with self.assertRaisesRegex(ValueError, 'must have a name'):
            d[b''] = Directory()
        with self.assertRaisesRegex(ValueError, 'must have a name'):
            d[b'/'] = Directory()

    def test_directory_set_wrong_type(self):
        d = Directory()
        with self.assertRaisesRegex(ValueError, 'Content or Directory'):
            d[b'entry'] = object()

    def test_directory_del_wrong_type(self):
        d = Directory()
        with self.assertRaisesRegex(ValueError, 'bytes Directory entry'):
            del d['foo']
        with self.assertRaisesRegex(ValueError, 'bytes Directory entry'):
            del d[42]



# Copyright (C) 2017 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.model import merkle


class MerkleTestNode(merkle.MerkleNode):
    type = 'tested_merkle_node_type'

    def __init__(self, data):
        super().__init__(data)
        self.compute_hash_called = 0

    def compute_hash(self):
        self.compute_hash_called += 1
        child_data = [
            child + b'=' + self[child].hash
            for child in sorted(self)
        ]

        return (
            b'hash('
            + b', '.join([self.data['value']] + child_data)
            + b')'
        )


class MerkleTestLeaf(merkle.MerkleLeaf):
    type = 'tested_merkle_leaf_type'

    def __init__(self, data):
        super().__init__(data)
        self.compute_hash_called = 0

    def compute_hash(self):
        self.compute_hash_called += 1
        return b'hash(' + self.data['value'] + b')'


class TestMerkleLeaf(unittest.TestCase):
    def setUp(self):
        self.data = {'value': b'value'}
        self.instance = MerkleTestLeaf(self.data)

    def test_equality(self):
        leaf1 = MerkleTestLeaf(self.data)
        leaf2 = MerkleTestLeaf(self.data)
        leaf3 = MerkleTestLeaf({})

        self.assertEqual(leaf1, leaf2)
        self.assertNotEqual(leaf1, leaf3)

    def test_hash(self):
        self.assertEqual(self.instance.compute_hash_called, 0)
        instance_hash = self.instance.hash
        self.assertEqual(self.instance.compute_hash_called, 1)
        instance_hash2 = self.instance.hash
        self.assertEqual(self.instance.compute_hash_called, 1)
        self.assertEqual(instance_hash, instance_hash2)

    def test_data(self):
        self.assertEqual(self.instance.get_data(), self.data)

    def test_collect(self):
        collected = self.instance.collect()
        self.assertEqual(
            collected, {
                self.instance.type: {
                    self.instance.hash: self.instance.get_data(),
                },
            },
        )
        collected2 = self.instance.collect()
        self.assertEqual(collected2, {})
        self.instance.reset_collect()
        collected3 = self.instance.collect()
        self.assertEqual(collected, collected3)

    def test_leaf(self):
        with self.assertRaisesRegex(ValueError, 'is a leaf'):
            self.instance[b'key1'] = 'Test'

        with self.assertRaisesRegex(ValueError, 'is a leaf'):
            del self.instance[b'key1']

        with self.assertRaisesRegex(ValueError, 'is a leaf'):
            self.instance[b'key1']

        with self.assertRaisesRegex(ValueError, 'is a leaf'):
            self.instance.update(self.data)


class TestMerkleNode(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.root = MerkleTestNode({'value': b'root'})
        self.nodes = {b'root': self.root}
        for i in (b'a', b'b', b'c'):
            value = b'root/' + i
            node = MerkleTestNode({
                'value': value,
            })
            self.root[i] = node
            self.nodes[value] = node
            for j in (b'a', b'b', b'c'):
                value2 = value + b'/' + j
                node2 = MerkleTestNode({
                    'value': value2,
                })
                node[j] = node2
                self.nodes[value2] = node2
                for k in (b'a', b'b', b'c'):
                    value3 = value2 + b'/' + j
                    node3 = MerkleTestNode({
                        'value': value3,
                    })
                    node2[j] = node3
                    self.nodes[value3] = node3

    def test_equality(self):
        node1 = merkle.MerkleNode({'foo': b'bar'})
        node2 = merkle.MerkleNode({'foo': b'bar'})
        node3 = merkle.MerkleNode({})

        self.assertEqual(node1, node2)
        self.assertNotEqual(node1, node3, node1 == node3)

        node1['foo'] = node3
        self.assertNotEqual(node1, node2)

        node2['foo'] = node3
        self.assertEqual(node1, node2)

    def test_hash(self):
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 0)

        # Root hash will compute hash for all the nodes
        hash = self.root.hash
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 1)
            self.assertIn(node.data['value'], hash)

        # Should use the cached value
        hash2 = self.root.hash
        self.assertEqual(hash, hash2)
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 1)

        # Should still use the cached value
        hash3 = self.root.update_hash(force=False)
        self.assertEqual(hash, hash3)
        for node in self.nodes.values():
            self.assertEqual(node.compute_hash_called, 1)

        # Force update of the cached value for a deeply nested node
        self.root[b'a'][b'b'].update_hash(force=True)
        for key, node in self.nodes.items():
            # update_hash rehashes all children
            if key.startswith(b'root/a/b'):
                self.assertEqual(node.compute_hash_called, 2)
            else:
                self.assertEqual(node.compute_hash_called, 1)

        hash4 = self.root.hash
        self.assertEqual(hash, hash4)
        for key, node in self.nodes.items():
            # update_hash also invalidates all parents
            if key in (b'root', b'root/a') or key.startswith(b'root/a/b'):
                self.assertEqual(node.compute_hash_called, 2)
            else:
                self.assertEqual(node.compute_hash_called, 1)

    def test_collect(self):
        collected = self.root.collect()
        self.assertEqual(len(collected[self.root.type]), len(self.nodes))
        for node in self.nodes.values():
            self.assertTrue(node.collected)
        collected2 = self.root.collect()
        self.assertEqual(collected2, {})

    def test_iter_tree(self):
        nodes = list(self.root.iter_tree())
        self.assertCountEqual(nodes, self.nodes.values())

    def test_get(self):
        for key in (b'a', b'b', b'c'):
            self.assertEqual(self.root[key], self.nodes[b'root/' + key])

        with self.assertRaisesRegex(KeyError, "b'nonexistent'"):
            self.root[b'nonexistent']

    def test_del(self):
        hash_root = self.root.hash
        hash_a = self.nodes[b'root/a'].hash
        del self.root[b'a'][b'c']
        hash_root2 = self.root.hash
        hash_a2 = self.nodes[b'root/a'].hash

        self.assertNotEqual(hash_root, hash_root2)
        self.assertNotEqual(hash_a, hash_a2)

        self.assertEqual(self.nodes[b'root/a/c'].parents, [])

        with self.assertRaisesRegex(KeyError, "b'nonexistent'"):
            del self.root[b'nonexistent']

    def test_update(self):
        hash_root = self.root.hash
        hash_b = self.root[b'b'].hash
        new_children = {
            b'c': MerkleTestNode({'value': b'root/b/new_c'}),
            b'd': MerkleTestNode({'value': b'root/b/d'}),
        }

        # collect all nodes
        self.root.collect()

        self.root[b'b'].update(new_children)

        # Ensure everyone got reparented
        self.assertEqual(new_children[b'c'].parents, [self.root[b'b']])
        self.assertEqual(new_children[b'd'].parents, [self.root[b'b']])
        self.assertEqual(self.nodes[b'root/b/c'].parents, [])

        hash_root2 = self.root.hash
        self.assertNotEqual(hash_root, hash_root2)
        self.assertIn(b'root/b/new_c', hash_root2)
        self.assertIn(b'root/b/d', hash_root2)

        hash_b2 = self.root[b'b'].hash
        self.assertNotEqual(hash_b, hash_b2)

        for key, node in self.nodes.items():
            if key in (b'root', b'root/b'):
                self.assertEqual(node.compute_hash_called, 2)
            else:
                self.assertEqual(node.compute_hash_called, 1)

        # Ensure we collected root, root/b, and both new children
        collected_after_update = self.root.collect()
        self.assertCountEqual(
            collected_after_update[MerkleTestNode.type],
            [self.nodes[b'root'].hash, self.nodes[b'root/b'].hash,
             new_children[b'c'].hash, new_children[b'd'].hash],
        )

        # test that noop updates doesn't invalidate anything
        self.root[b'a'][b'b'].update({})
        self.assertEqual(self.root.collect(), {})



# Copyright (C) 2018-2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import tarfile
import tempfile
import unittest

from click.testing import CliRunner
import pytest

from swh.model import cli
from swh.model.hashutil import hash_to_hex
from swh.model.tests.test_from_disk import DataMixin


@pytest.mark.fs
class TestIdentify(DataMixin, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.runner = CliRunner()

    def assertPidOK(self, result, pid):
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output.split()[0], pid)

    def test_content_id(self):
        """identify file content"""
        self.make_contents(self.tmpdir_name)
        for filename, content in self.contents.items():
            path = os.path.join(self.tmpdir_name, filename)
            result = self.runner.invoke(cli.identify,
                                        ['--type', 'content', path])
            self.assertPidOK(result,
                             'swh:1:cnt:' + hash_to_hex(content['sha1_git']))

    def test_content_id_from_stdin(self):
        """identify file content"""
        self.make_contents(self.tmpdir_name)
        for _, content in self.contents.items():
            result = self.runner.invoke(cli.identify,
                                        input=content['data'])
            self.assertPidOK(result,
                             'swh:1:cnt:' + hash_to_hex(content['sha1_git']))

    def test_directory_id(self):
        """identify an entire directory"""
        self.make_from_tarball(self.tmpdir_name)
        path = os.path.join(self.tmpdir_name, b'sample-folder')
        result = self.runner.invoke(cli.identify,
                                    ['--type', 'directory', path])
        self.assertPidOK(result,
                         'swh:1:dir:e8b0f1466af8608c8a3fb9879db172b887e80759')

    def test_snapshot_id(self):
        """identify a snapshot"""
        tarball = os.path.join(os.path.dirname(__file__), 'data', 'repos',
                               'sample-repo.tgz')
        with tempfile.TemporaryDirectory(prefix='swh.model.cli') as d:
            with tarfile.open(tarball, 'r:gz') as t:
                t.extractall(d)
                repo_dir = os.path.join(d, 'sample-repo')
                result = self.runner.invoke(cli.identify,
                                            ['--type', 'snapshot', repo_dir])
                self.assertPidOK(
                    result,
                    'swh:1:snp:abc888898124270905a0ef3c67e872ce08e7e0c1')

    def test_origin_id(self):
        """identify an origin URL"""
        url = 'https://github.com/torvalds/linux'
        result = self.runner.invoke(cli.identify, ['--type', 'origin', url])
        self.assertPidOK(result,
                         'swh:1:ori:b63a575fe3faab7692c9f38fb09d4bb45651bb0f')

    def test_symlink(self):
        """identify symlink --- both itself and target"""
        regular = os.path.join(self.tmpdir_name, b'foo.txt')
        link = os.path.join(self.tmpdir_name, b'bar.txt')
        open(regular, 'w').write('foo\n')
        os.symlink(os.path.basename(regular), link)

        result = self.runner.invoke(cli.identify, [link])
        self.assertPidOK(result,
                         'swh:1:cnt:257cc5642cb1a054f08cc83f2d943e56fd3ebe99')

        result = self.runner.invoke(cli.identify, ['--no-dereference', link])
        self.assertPidOK(result,
                         'swh:1:cnt:996f1789ff67c0e3f69ef5933a55d54c5d0e9954')

    def test_show_filename(self):
        """filename is shown by default"""
        self.make_contents(self.tmpdir_name)
        for filename, content in self.contents.items():
            path = os.path.join(self.tmpdir_name, filename)
            result = self.runner.invoke(cli.identify,
                                        ['--type', 'content', path])

            self.assertEqual(result.exit_code, 0)
            self.assertEqual(result.output.rstrip(),
                             'swh:1:cnt:%s\t%s' %
                             (hash_to_hex(content['sha1_git']), path.decode()))

    def test_hide_filename(self):
        """filename is hidden upon request"""
        self.make_contents(self.tmpdir_name)
        for filename, content in self.contents.items():
            path = os.path.join(self.tmpdir_name, filename)
            result = self.runner.invoke(cli.identify,
                                        ['--type', 'content', '--no-filename',
                                         path])
            self.assertPidOK(result,
                             'swh:1:cnt:' + hash_to_hex(content['sha1_git']))

    def test_auto_content(self):
        """automatic object type detection: content"""
        with tempfile.NamedTemporaryFile(prefix='swh.model.cli') as f:
            result = self.runner.invoke(cli.identify, [f.name])
            self.assertEqual(result.exit_code, 0)
            self.assertRegex(result.output, r'^swh:\d+:cnt:')

    def test_auto_directory(self):
        """automatic object type detection: directory"""
        with tempfile.TemporaryDirectory(prefix='swh.model.cli') as dirname:
            result = self.runner.invoke(cli.identify, [dirname])
            self.assertEqual(result.exit_code, 0)
            self.assertRegex(result.output, r'^swh:\d+:dir:')

    def test_auto_origin(self):
        """automatic object type detection: origin"""
        result = self.runner.invoke(cli.identify,
                                    ['https://github.com/torvalds/linux'])
        self.assertEqual(result.exit_code, 0)
        self.assertRegex(result.output, r'^swh:\d+:ori:')

    def test_verify_content(self):
        """identifier verification"""
        self.make_contents(self.tmpdir_name)
        for filename, content in self.contents.items():
            expected_id = 'swh:1:cnt:' + hash_to_hex(content['sha1_git'])

            # match
            path = os.path.join(self.tmpdir_name, filename)
            result = self.runner.invoke(cli.identify,
                                        ['--verify', expected_id, path])
            self.assertEqual(result.exit_code, 0)

            # mismatch
            with open(path, 'a') as f:
                f.write('trailing garbage to make verification fail')
            result = self.runner.invoke(cli.identify,
                                        ['--verify', expected_id, path])
            self.assertEqual(result.exit_code, 1)



# Copyright (C) 2018-2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Dict, List, Optional, Union

import attr
from attrs_strict import type_validator
import dateutil.parser
import iso8601

from .identifiers import (
    normalize_timestamp, directory_identifier, revision_identifier,
    release_identifier, snapshot_identifier
)
from .hashutil import DEFAULT_ALGORITHMS, hash_to_bytes, MultiHash


class MissingData(Exception):
    """Raised by `Content.with_data` when it has no way of fetching the
    data (but not when fetching the data fails)."""
    pass


SHA1_SIZE = 20

# TODO: Limit this to 20 bytes
Sha1Git = bytes


def dictify(value):
    "Helper function used by BaseModel.to_dict()"
    if isinstance(value, BaseModel):
        return value.to_dict()
    elif isinstance(value, Enum):
        return value.value
    elif isinstance(value, dict):
        return {k: dictify(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [dictify(v) for v in value]
    else:
        return value


class BaseModel:
    """Base class for SWH model classes.

    Provides serialization/deserialization to/from Python dictionaries,
    that are suitable for JSON/msgpack-like formats."""

    def to_dict(self):
        """Wrapper of `attr.asdict` that can be overridden by subclasses
        that have special handling of some of the fields."""
        return dictify(attr.asdict(self, recurse=False))

    @classmethod
    def from_dict(cls, d):
        """Takes a dictionary representing a tree of SWH objects, and
        recursively builds the corresponding objects."""
        return cls(**d)


class HashableObject(metaclass=ABCMeta):
    """Mixin to automatically compute object identifier hash when
    the associated model is instantiated."""

    @staticmethod
    @abstractmethod
    def compute_hash(object_dict):
        """Derived model classes must implement this to compute
        the object hash from its dict representation."""
        pass

    def __attrs_post_init__(self):
        if not self.id:
            obj_id = hash_to_bytes(self.compute_hash(self.to_dict()))
            object.__setattr__(self, 'id', obj_id)


@attr.s(frozen=True)
class Person(BaseModel):
    """Represents the author/committer of a revision or release."""
    fullname = attr.ib(
        type=bytes,
        validator=type_validator())
    name = attr.ib(
        type=Optional[bytes],
        validator=type_validator())
    email = attr.ib(
        type=Optional[bytes],
        validator=type_validator())

    @classmethod
    def from_fullname(cls, fullname: bytes):
        """Returns a Person object, by guessing the name and email from the
        fullname, in the `name <email>` format.

        The fullname is left unchanged."""
        if fullname is None:
            raise TypeError('fullname is None.')

        name: Optional[bytes]
        email: Optional[bytes]

        try:
            open_bracket = fullname.index(b'<')
        except ValueError:
            name = fullname
            email = None
        else:
            raw_name = fullname[:open_bracket]
            raw_email = fullname[open_bracket+1:]

            if not raw_name:
                name = None
            else:
                name = raw_name.strip()

            try:
                close_bracket = raw_email.rindex(b'>')
            except ValueError:
                email = raw_email
            else:
                email = raw_email[:close_bracket]

        return Person(
            name=name or None,
            email=email or None,
            fullname=fullname,
        )


@attr.s(frozen=True)
class Timestamp(BaseModel):
    """Represents a naive timestamp from a VCS."""
    seconds = attr.ib(
        type=int,
        validator=type_validator())
    microseconds = attr.ib(
        type=int,
        validator=type_validator())

    @seconds.validator
    def check_seconds(self, attribute, value):
        """Check that seconds fit in a 64-bits signed integer."""
        if not (-2**63 <= value < 2**63):
            raise ValueError('Seconds must be a signed 64-bits integer.')

    @microseconds.validator
    def check_microseconds(self, attribute, value):
        """Checks that microseconds are positive and < 1000000."""
        if not (0 <= value < 10**6):
            raise ValueError('Microseconds must be in [0, 1000000[.')


@attr.s(frozen=True)
class TimestampWithTimezone(BaseModel):
    """Represents a TZ-aware timestamp from a VCS."""
    timestamp = attr.ib(
        type=Timestamp,
        validator=type_validator())
    offset = attr.ib(
        type=int,
        validator=type_validator())
    negative_utc = attr.ib(
        type=bool,
        validator=type_validator())

    @offset.validator
    def check_offset(self, attribute, value):
        """Checks the offset is a 16-bits signed integer (in theory, it
        should always be between -14 and +14 hours)."""
        if not (-2**15 <= value < 2**15):
            # max 14 hours offset in theory, but you never know what
            # you'll find in the wild...
            raise ValueError('offset too large: %d minutes' % value)

    @negative_utc.validator
    def check_negative_utc(self, attribute, value):
        if self.offset and value:
            raise ValueError("negative_utc can only be True is offset=0")

    @classmethod
    def from_dict(cls, obj: Union[Dict, datetime.datetime, int]):
        """Builds a TimestampWithTimezone from any of the formats
        accepted by :func:`swh.model.normalize_timestamp`."""
        # TODO: this accept way more types than just dicts; find a better
        # name
        d = normalize_timestamp(obj)
        return cls(
            timestamp=Timestamp.from_dict(d['timestamp']),
            offset=d['offset'],
            negative_utc=d['negative_utc'])

    @classmethod
    def from_datetime(cls, dt: datetime.datetime):
        return cls.from_dict(dt)

    @classmethod
    def from_iso8601(cls, s):
        """Builds a TimestampWithTimezone from an ISO8601-formatted string.
        """
        dt = iso8601.parse_date(s)
        tstz = cls.from_datetime(dt)
        if dt.tzname() == '-00:00':
            tstz = attr.evolve(tstz, negative_utc=True)
        return tstz


@attr.s(frozen=True)
class Origin(BaseModel):
    """Represents a software source: a VCS and an URL."""
    url = attr.ib(
        type=str,
        validator=type_validator())


@attr.s(frozen=True)
class OriginVisit(BaseModel):
    """Represents a visit of an origin at a given point in time, by a
    SWH loader."""
    origin = attr.ib(
        type=str,
        validator=type_validator())
    date = attr.ib(
        type=datetime.datetime,
        validator=type_validator())
    status = attr.ib(
        type=str,
        validator=attr.validators.in_(['ongoing', 'full', 'partial']))
    type = attr.ib(
        type=str,
        validator=type_validator())
    snapshot = attr.ib(
        type=Optional[Sha1Git],
        validator=type_validator())
    metadata = attr.ib(
        type=Optional[Dict[str, object]],
        validator=type_validator(),
        default=None)
    visit = attr.ib(
        type=Optional[int],
        validator=type_validator(),
        default=None)
    """Should not be set before calling 'origin_visit_add()'."""

    def to_dict(self):
        """Serializes the date as a string and omits the visit id if it is
        `None`."""
        ov = super().to_dict()
        if ov['visit'] is None:
            del ov['visit']
        return ov

    @classmethod
    def from_dict(cls, d):
        """Parses the date from a string, and accepts missing visit ids."""
        if isinstance(d['date'], str):
            d = d.copy()
            d['date'] = dateutil.parser.parse(d['date'])
        return super().from_dict(d)


@attr.s(frozen=True)
class OriginVisitUpdate(BaseModel):
    """Represents a visit update of an origin at a given point in time.

    """
    origin = attr.ib(
        type=str,
        validator=type_validator())
    visit = attr.ib(
        type=int,
        validator=type_validator())

    date = attr.ib(
        type=datetime.datetime,
        validator=type_validator())
    status = attr.ib(
        type=str,
        validator=attr.validators.in_(['ongoing', 'full', 'partial']))
    snapshot = attr.ib(
        type=Optional[Sha1Git],
        validator=type_validator())
    metadata = attr.ib(
        type=Optional[Dict[str, object]],
        validator=type_validator(),
        default=None)


class TargetType(Enum):
    """The type of content pointed to by a snapshot branch. Usually a
    revision or an alias."""
    CONTENT = 'content'
    DIRECTORY = 'directory'
    REVISION = 'revision'
    RELEASE = 'release'
    SNAPSHOT = 'snapshot'
    ALIAS = 'alias'


class ObjectType(Enum):
    """The type of content pointed to by a release. Usually a revision"""
    CONTENT = 'content'
    DIRECTORY = 'directory'
    REVISION = 'revision'
    RELEASE = 'release'
    SNAPSHOT = 'snapshot'


@attr.s(frozen=True)
class SnapshotBranch(BaseModel):
    """Represents one of the branches of a snapshot."""
    target = attr.ib(
        type=bytes,
        validator=type_validator())
    target_type = attr.ib(
        type=TargetType,
        validator=type_validator())

    @target.validator
    def check_target(self, attribute, value):
        """Checks the target type is not an alias, checks the target is a
        valid sha1_git."""
        if self.target_type != TargetType.ALIAS and self.target is not None:
            if len(value) != 20:
                raise ValueError('Wrong length for bytes identifier: %d' %
                                 len(value))

    @classmethod
    def from_dict(cls, d):
        return cls(
            target=d['target'],
            target_type=TargetType(d['target_type']))


@attr.s(frozen=True)
class Snapshot(BaseModel, HashableObject):
    """Represents the full state of an origin at a given point in time."""
    branches = attr.ib(
        type=Dict[bytes, Optional[SnapshotBranch]],
        validator=type_validator())
    id = attr.ib(
        type=Sha1Git,
        validator=type_validator(),
        default=b'')

    @staticmethod
    def compute_hash(object_dict):
        return snapshot_identifier(object_dict)

    @classmethod
    def from_dict(cls, d):
        d = d.copy()
        return cls(
            branches={
                name: SnapshotBranch.from_dict(branch) if branch else None
                for (name, branch) in d.pop('branches').items()
            },
            **d)


@attr.s(frozen=True)
class Release(BaseModel, HashableObject):
    name = attr.ib(
        type=bytes,
        validator=type_validator())
    message = attr.ib(
        type=Optional[bytes],
        validator=type_validator())
    target = attr.ib(
        type=Optional[Sha1Git],
        validator=type_validator())
    target_type = attr.ib(
        type=ObjectType,
        validator=type_validator())
    synthetic = attr.ib(
        type=bool,
        validator=type_validator())
    author = attr.ib(
        type=Optional[Person],
        validator=type_validator(),
        default=None)
    date = attr.ib(
        type=Optional[TimestampWithTimezone],
        validator=type_validator(),
        default=None)
    metadata = attr.ib(
        type=Optional[Dict[str, object]],
        validator=type_validator(),
        default=None)
    id = attr.ib(
        type=Sha1Git,
        validator=type_validator(),
        default=b'')

    @staticmethod
    def compute_hash(object_dict):
        return release_identifier(object_dict)

    @author.validator
    def check_author(self, attribute, value):
        """If the author is `None`, checks the date is `None` too."""
        if self.author is None and self.date is not None:
            raise ValueError('release date must be None if author is None.')

    def to_dict(self):
        rel = super().to_dict()
        if rel['metadata'] is None:
            del rel['metadata']
        return rel

    @classmethod
    def from_dict(cls, d):
        d = d.copy()
        if d.get('author'):
            d['author'] = Person.from_dict(d['author'])
        if d.get('date'):
            d['date'] = TimestampWithTimezone.from_dict(d['date'])
        return cls(
            target_type=ObjectType(d.pop('target_type')),
            **d)


class RevisionType(Enum):
    GIT = 'git'
    TAR = 'tar'
    DSC = 'dsc'
    SUBVERSION = 'svn'
    MERCURIAL = 'hg'


@attr.s(frozen=True)
class Revision(BaseModel, HashableObject):
    message = attr.ib(
        type=bytes,
        validator=type_validator())
    author = attr.ib(
        type=Person,
        validator=type_validator())
    committer = attr.ib(
        type=Person,
        validator=type_validator())
    date = attr.ib(
        type=Optional[TimestampWithTimezone],
        validator=type_validator())
    committer_date = attr.ib(
        type=Optional[TimestampWithTimezone],
        validator=type_validator())
    type = attr.ib(
        type=RevisionType,
        validator=type_validator())
    directory = attr.ib(
        type=Sha1Git,
        validator=type_validator())
    synthetic = attr.ib(
        type=bool,
        validator=type_validator())
    metadata = attr.ib(
        type=Optional[Dict[str, object]],
        validator=type_validator(),
        default=None)
    parents = attr.ib(
        type=List[Sha1Git],
        validator=type_validator(),
        default=attr.Factory(list))
    id = attr.ib(
        type=Sha1Git,
        validator=type_validator(),
        default=b'')

    @staticmethod
    def compute_hash(object_dict):
        return revision_identifier(object_dict)

    @classmethod
    def from_dict(cls, d):
        d = d.copy()
        date = d.pop('date')
        if date:
            date = TimestampWithTimezone.from_dict(date)

        committer_date = d.pop('committer_date')
        if committer_date:
            committer_date = TimestampWithTimezone.from_dict(
                committer_date)

        return cls(
            author=Person.from_dict(d.pop('author')),
            committer=Person.from_dict(d.pop('committer')),
            date=date,
            committer_date=committer_date,
            type=RevisionType(d.pop('type')),
            **d)


@attr.s(frozen=True)
class DirectoryEntry(BaseModel):
    name = attr.ib(
        type=bytes,
        validator=type_validator())
    type = attr.ib(
        type=str,
        validator=attr.validators.in_(['file', 'dir', 'rev']))
    target = attr.ib(
        type=Sha1Git,
        validator=type_validator())
    perms = attr.ib(
        type=int,
        validator=type_validator())
    """Usually one of the values of `swh.model.from_disk.DentryPerms`."""


@attr.s(frozen=True)
class Directory(BaseModel, HashableObject):
    entries = attr.ib(
        type=List[DirectoryEntry],
        validator=type_validator())
    id = attr.ib(
        type=Sha1Git,
        validator=type_validator(),
        default=b'')

    @staticmethod
    def compute_hash(object_dict):
        return directory_identifier(object_dict)

    @classmethod
    def from_dict(cls, d):
        d = d.copy()
        return cls(
            entries=[DirectoryEntry.from_dict(entry)
                     for entry in d.pop('entries')],
            **d)


@attr.s(frozen=True)
class BaseContent(BaseModel):
    status = attr.ib(
        type=str,
        validator=attr.validators.in_(['visible', 'hidden', 'absent']))

    @staticmethod
    def _hash_data(data: bytes):
        """Hash some data, returning most of the fields of a content object"""
        d = MultiHash.from_data(data).digest()
        d['data'] = data
        d['length'] = len(data)

        return d

    @classmethod
    def from_dict(cls, d, use_subclass=True):
        if use_subclass:
            # Chooses a subclass to instantiate instead.
            if d['status'] == 'absent':
                return SkippedContent.from_dict(d)
            else:
                return Content.from_dict(d)
        else:
            return super().from_dict(d)

    def get_hash(self, hash_name):
        if hash_name not in DEFAULT_ALGORITHMS:
            raise ValueError('{} is not a valid hash name.'.format(hash_name))
        return getattr(self, hash_name)

    def hashes(self) -> Dict[str, bytes]:
        """Returns a dictionary {hash_name: hash_value}"""
        return {algo: getattr(self, algo) for algo in DEFAULT_ALGORITHMS}


@attr.s(frozen=True)
class Content(BaseContent):
    sha1 = attr.ib(
        type=bytes,
        validator=type_validator())
    sha1_git = attr.ib(
        type=Sha1Git,
        validator=type_validator())
    sha256 = attr.ib(
        type=bytes,
        validator=type_validator())
    blake2s256 = attr.ib(
        type=bytes,
        validator=type_validator())

    length = attr.ib(
        type=int,
        validator=type_validator())

    status = attr.ib(
        type=str,
        validator=attr.validators.in_(['visible', 'hidden']),
        default='visible')

    data = attr.ib(
        type=Optional[bytes],
        validator=type_validator(),
        default=None)

    ctime = attr.ib(
        type=Optional[datetime.datetime],
        validator=type_validator(),
        default=None)

    @length.validator
    def check_length(self, attribute, value):
        """Checks the length is positive."""
        if value < 0:
            raise ValueError('Length must be positive.')

    def to_dict(self):
        content = super().to_dict()
        if content['data'] is None:
            del content['data']
        return content

    @classmethod
    def from_data(cls, data, status='visible', ctime=None) -> 'Content':
        """Generate a Content from a given `data` byte string.

        This populates the Content with the hashes and length for the data
        passed as argument, as well as the data itself.
        """
        d = cls._hash_data(data)
        d['status'] = status
        d['ctime'] = ctime
        return cls(**d)

    @classmethod
    def from_dict(cls, d):
        if isinstance(d.get('ctime'), str):
            d = d.copy()
            d['ctime'] = dateutil.parser.parse(d['ctime'])
        return super().from_dict(d, use_subclass=False)

    def with_data(self) -> 'Content':
        """Loads the `data` attribute; meaning that it is guaranteed not to
        be None after this call.

        This call is almost a no-op, but subclasses may overload this method
        to lazy-load data (eg. from disk or objstorage)."""
        if self.data is None:
            raise MissingData('Content data is None.')
        return self


@attr.s(frozen=True)
class SkippedContent(BaseContent):
    sha1 = attr.ib(
        type=Optional[bytes],
        validator=type_validator())
    sha1_git = attr.ib(
        type=Optional[Sha1Git],
        validator=type_validator())
    sha256 = attr.ib(
        type=Optional[bytes],
        validator=type_validator())
    blake2s256 = attr.ib(
        type=Optional[bytes],
        validator=type_validator())

    length = attr.ib(
        type=Optional[int],
        validator=type_validator())

    status = attr.ib(
        type=str,
        validator=attr.validators.in_(['absent']))
    reason = attr.ib(
        type=Optional[str],
        validator=type_validator(),
        default=None)

    origin = attr.ib(
        type=Optional[str],
        validator=type_validator(),
        default=None)

    ctime = attr.ib(
        type=Optional[datetime.datetime],
        validator=type_validator(),
        default=None)

    @reason.validator
    def check_reason(self, attribute, value):
        """Checks the reason is full if status != absent."""
        assert self.reason == value
        if value is None:
            raise ValueError('Must provide a reason if content is absent.')

    @length.validator
    def check_length(self, attribute, value):
        """Checks the length is positive or -1."""
        if value < -1:
            raise ValueError('Length must be positive or -1.')

    def to_dict(self):
        content = super().to_dict()
        if content['origin'] is None:
            del content['origin']
        return content

    @classmethod
    def from_data(
            cls,
            data: bytes,
            reason: str,
            ctime: Optional[datetime.datetime] = None) -> 'SkippedContent':
        """Generate a SkippedContent from a given `data` byte string.

        This populates the SkippedContent with the hashes and length for the
        data passed as argument.

        You can use `attr.evolve` on such a generated content to nullify some
        of its attributes, e.g. for tests.
        """
        d = cls._hash_data(data)
        del d['data']
        d['status'] = 'absent'
        d['reason'] = reason
        d['ctime'] = ctime
        return cls(**d)

    @classmethod
    def from_dict(cls, d):
        d2 = d.copy()
        if d2.pop('data', None) is not None:
            raise ValueError('SkippedContent has no "data" attribute %r' % d)
        return super().from_dict(d2, use_subclass=False)
